# This script generates a Python interface for an Apple Macintosh Manager.
# It uses the "bgen" package to generate C code.
# The function specifications are generated by scanning the mamager's header file,
# using the "scantools" package (customized for this particular manager).

import string

# Declarations that change for each manager
MACHEADERFILE = 'Components.h'          # The Apple header file
MODNAME = '_Cm'                         # The name of the module

# The following is *usually* unchanged but may still require tuning
MODPREFIX = 'Cm'                        # The prefix for module-wide routines
C_OBJECTPREFIX = 'CmpObj'       # The prefix for object methods
CI_OBJECTPREFIX = 'CmpInstObj'
INPUTFILE = string.lower(MODPREFIX) + 'gen.py' # The file generated by the scanner
OUTPUTFILE = MODNAME + "module.c"       # The file generated by this program

from macsupport import *

# Create the type objects

includestuff = includestuff + """
#include <Carbon/Carbon.h>

#ifdef USE_TOOLBOX_OBJECT_GLUE
extern PyObject *_CmpObj_New(Component);
extern int _CmpObj_Convert(PyObject *, Component *);
extern PyObject *_CmpInstObj_New(ComponentInstance);
extern int _CmpInstObj_Convert(PyObject *, ComponentInstance *);

#define CmpObj_New _CmpObj_New
#define CmpObj_Convert _CmpObj_Convert
#define CmpInstObj_New _CmpInstObj_New
#define CmpInstObj_Convert _CmpInstObj_Convert
#endif

/*
** Parse/generate ComponentDescriptor records
*/
static PyObject *
CmpDesc_New(ComponentDescription *itself)
{

        return Py_BuildValue("O&O&O&ll",
                PyMac_BuildOSType, itself->componentType,
                PyMac_BuildOSType, itself->componentSubType,
                PyMac_BuildOSType, itself->componentManufacturer,
                itself->componentFlags, itself->componentFlagsMask);
}

static int
CmpDesc_Convert(PyObject *v, ComponentDescription *p_itself)
{
        return PyArg_ParseTuple(v, "O&O&O&ll",
                PyMac_GetOSType, &p_itself->componentType,
                PyMac_GetOSType, &p_itself->componentSubType,
                PyMac_GetOSType, &p_itself->componentManufacturer,
                &p_itself->componentFlags, &p_itself->componentFlagsMask);
}

"""

initstuff = initstuff + """
        PyMac_INIT_TOOLBOX_OBJECT_NEW(Component, CmpObj_New);
        PyMac_INIT_TOOLBOX_OBJECT_CONVERT(Component, CmpObj_Convert);
        PyMac_INIT_TOOLBOX_OBJECT_NEW(ComponentInstance, CmpInstObj_New);
        PyMac_INIT_TOOLBOX_OBJECT_CONVERT(ComponentInstance, CmpInstObj_Convert);
"""

ComponentDescription = OpaqueType('ComponentDescription', 'CmpDesc')
Component = OpaqueByValueType('Component', C_OBJECTPREFIX)
ComponentInstance = OpaqueByValueType('ComponentInstance', CI_OBJECTPREFIX)
ComponentResult = Type("ComponentResult", "l")

ComponentResourceHandle = OpaqueByValueType("ComponentResourceHandle", "ResObj")

class MyCIObjectDefinition(PEP253Mixin, GlobalObjectDefinition):
    def outputCheckNewArg(self):
        Output("""if (itself == NULL) {
                                PyErr_SetString(Cm_Error,"NULL ComponentInstance");
                                return NULL;
                        }""")

class MyCObjectDefinition(PEP253Mixin, GlobalObjectDefinition):
    def outputCheckNewArg(self):
        Output("""if (itself == NULL) {
                                /* XXXX Or should we return None? */
                                PyErr_SetString(Cm_Error,"No such component");
                                return NULL;
                        }""")

    def outputCheckConvertArg(self):
        Output("""if ( v == Py_None ) {
                                *p_itself = 0;
                                return 1;
        }""")

# Create the generator groups and link them
module = MacModule(MODNAME, MODPREFIX, includestuff, finalstuff, initstuff)
ci_object = MyCIObjectDefinition('ComponentInstance', CI_OBJECTPREFIX,
                'ComponentInstance')
c_object = MyCObjectDefinition('Component', C_OBJECTPREFIX, 'Component')
module.addobject(ci_object)
module.addobject(c_object)

# Create the generator classes used to populate the lists
Function = OSErrWeakLinkFunctionGenerator
Method = OSErrWeakLinkMethodGenerator

# Create and populate the lists
functions = []
c_methods = []
ci_methods = []
execfile(INPUTFILE)

# add the populated lists to the generator groups
# (in a different wordl the scan program would generate this)
for f in functions: module.add(f)
for f in c_methods: c_object.add(f)
for f in ci_methods: ci_object.add(f)

# generate output (open the output file as late as possible)
SetOutputFileName(OUTPUTFILE)
module.generate()



# Module 'dump'
#
# Print python code that reconstructs a variable.
# This only works in certain cases.
#
# It works fine for:
# - ints and floats (except NaNs and other weird things)
# - strings
# - compounds and lists, provided it works for all their elements
# - imported modules, provided their name is the module name
#
# It works for top-level dictionaries but not for dictionaries
# contained in other objects (could be made to work with some hassle
# though).
#
# It does not work for functions (all sorts), classes, class objects,
# windows, files etc.
#
# Finally, objects referenced by more than one name or contained in more
# than one other object lose their sharing property (this is bad for
# strings used as exception identifiers, for instance).

# Dump a whole symbol table
#
def dumpsymtab(dict):
    for key in dict.keys():
        dumpvar(key, dict[key])

# Dump a single variable
#
def dumpvar(name, x):
    import sys
    t = type(x)
    if t == type({}):
        print name, '= {}'
        for key in x.keys():
            item = x[key]
            if not printable(item):
                print '#',
            print name, '[', `key`, '] =', `item`
    elif t in (type(''), type(0), type(0.0), type([]), type(())):
        if not printable(x):
            print '#',
        print name, '=', `x`
    elif t == type(sys):
        print 'import', name, '#', x
    else:
        print '#', name, '=', x

# check if a value is printable in a way that can be read back with input()
#
def printable(x):
    t = type(x)
    if t in (type(''), type(0), type(0.0)):
        return 1
    if t in (type([]), type(())):
        for item in x:
            if not printable(item):
                return 0
        return 1
    if x == {}:
        return 1
    return 0



# This script generates a Python interface for an Apple Macintosh Manager.
# It uses the "bgen" package to generate C code.
# The function specifications are generated by scanning the mamager's header file,
# using the "scantools" package (customized for this particular manager).

import string

# Declarations that change for each manager
MACHEADERFILE = 'Lists.h'               # The Apple header file
MODNAME = '_List'                               # The name of the module
OBJECTNAME = 'List'                     # The basic name of the objects used here
KIND = 'Handle'                         # Usually 'Ptr' or 'Handle'

# The following is *usually* unchanged but may still require tuning
MODPREFIX = 'List'                      # The prefix for module-wide routines
OBJECTTYPE = "ListHandle"               # The C type used to represent them
OBJECTPREFIX = MODPREFIX + 'Obj'        # The prefix for object methods
INPUTFILE = string.lower(MODPREFIX) + 'gen.py' # The file generated by the scanner
OUTPUTFILE = MODNAME + "module.c"       # The file generated by this program

from macsupport import *

# Create the type objects
ListHandle = OpaqueByValueType("ListHandle", "ListObj")
ListRef = ListHandle # Obsolete, but used in Lists.h
Cell = Point
ListBounds = Rect
ListBounds_ptr = Rect_ptr

ListDefSpec = ListDefSpec_ptr = OpaqueType("ListDefSpec", "PyMac_BuildListDefSpec", "PyMac_GetListDefSpec")

VarOutBufferShortsize = VarHeapOutputBufferType('char', 'short', 's')   # (buf, &len)
InBufferShortsize = VarInputBufferType('char', 'short', 's')            # (buf, len)

RgnHandle = OpaqueByValueType("RgnHandle", "ResObj")
DataHandle = OpaqueByValueType("DataHandle", "ResObj")
Handle = OpaqueByValueType("Handle", "ResObj")
CGrafPtr = OpaqueByValueType("CGrafPtr", "GrafObj")
EventModifiers = Type("EventModifiers", "H")

includestuff = includestuff + """
#include <Carbon/Carbon.h>

#ifdef USE_TOOLBOX_OBJECT_GLUE
extern PyObject *_ListObj_New(ListHandle);
extern int _ListObj_Convert(PyObject *, ListHandle *);

#define ListObj_New _ListObj_New
#define ListObj_Convert _ListObj_Convert
#endif

#define as_List(x) ((ListHandle)x)
#define as_Resource(lh) ((Handle)lh)

static ListDefUPP myListDefFunctionUPP;

"""

initstuff = initstuff + """
myListDefFunctionUPP = NewListDefUPP((ListDefProcPtr)myListDefFunction);

PyMac_INIT_TOOLBOX_OBJECT_NEW(ListHandle, ListObj_New);
PyMac_INIT_TOOLBOX_OBJECT_CONVERT(ListHandle, ListObj_Convert);
"""

class ListMethodGenerator(MethodGenerator):
    """Similar to MethodGenerator, but has self as last argument"""

    def parseArgumentList(self, args):
        args, a0 = args[:-1], args[-1]
        t0, n0, m0 = a0
        if m0 != InMode:
            raise ValueError, "method's 'self' must be 'InMode'"
        self.itself = Variable(t0, "_self->ob_itself", SelfMode)
        FunctionGenerator.parseArgumentList(self, args)
        self.argumentList.append(self.itself)

class MyObjectDefinition(PEP253Mixin, GlobalObjectDefinition):
    # XXXX Should inherit from Resource
    getsetlist = [(
            'listFlags',
            'return Py_BuildValue("l", (long)GetListFlags(self->ob_itself) & 0xff);',
            'if (!PyArg_Parse(v, "B", &(*self->ob_itself)->listFlags)) return -1;',
            None,
            ), (
            'selFlags',
            'return Py_BuildValue("l", (long)GetListSelectionFlags(self->ob_itself) & 0xff);',
            'if (!PyArg_Parse(v, "B", &(*self->ob_itself)->selFlags)) return -1;',
            None,
            ), (
            'cellSize',
            'return Py_BuildValue("O&", PyMac_BuildPoint, (*self->ob_itself)->cellSize);',
            'if (!PyArg_Parse(v, "O&", PyMac_GetPoint, &(*self->ob_itself)->cellSize)) return -1;',
            None
            )]

    def outputStructMembers(self):
        ObjectDefinition.outputStructMembers(self)
        Output("PyObject *ob_ldef_func;")
        Output("int ob_must_be_disposed;")

    def outputCheckNewArg(self):
        Output("""if (itself == NULL) {
                                PyErr_SetString(List_Error,"Cannot create null List");
                                return NULL;
                        }""")

    def outputInitStructMembers(self):
        ObjectDefinition.outputInitStructMembers(self)
        Output("it->ob_ldef_func = NULL;")
        Output("it->ob_must_be_disposed = 1;")
        Output("SetListRefCon(itself, (long)it);")

    def outputFreeIt(self, itselfname):
        Output("Py_XDECREF(self->ob_ldef_func);")
        Output("self->ob_ldef_func = NULL;")
        Output("SetListRefCon(self->ob_itself, (long)0);")
        Output("if (self->ob_must_be_disposed && %s) LDispose(%s);", itselfname, itselfname)

# From here on it's basically all boiler plate...

finalstuff = finalstuff + """
static void myListDefFunction(SInt16 message,
                       Boolean selected,
                       Rect *cellRect,
                       Cell theCell,
                       SInt16 dataOffset,
                       SInt16 dataLen,
                       ListHandle theList)
{
        PyObject *listDefFunc, *args, *rv=NULL;
        ListObject *self;

        self = (ListObject*)GetListRefCon(theList);
        if (self == NULL || self->ob_itself != theList)
                return;  /* nothing we can do */
        listDefFunc = self->ob_ldef_func;
        if (listDefFunc == NULL)
                return;  /* nothing we can do */
        args = Py_BuildValue("hbO&O&hhO", message,
                                          selected,
                                          PyMac_BuildRect, cellRect,
                                          PyMac_BuildPoint, theCell,
                                          dataOffset,
                                          dataLen,
                                          self);
        if (args != NULL) {
                rv = PyEval_CallObject(listDefFunc, args);
                Py_DECREF(args);
        }
        if (rv == NULL) {
                PySys_WriteStderr("error in list definition callback:\\n");
                PyErr_Print();
        } else {
                Py_DECREF(rv);
        }
}
"""

# Create the generator groups and link them
module = MacModule(MODNAME, MODPREFIX, includestuff, finalstuff, initstuff)
object = MyObjectDefinition(OBJECTNAME, OBJECTPREFIX, OBJECTTYPE)
module.addobject(object)

# Create the generator classes used to populate the lists
Function = FunctionGenerator
Method = ListMethodGenerator

# Create and populate the lists
functions = []
methods = []
execfile(INPUTFILE)

# Function to convert any handle to a list and vv.
##f = Function(ListHandle, 'as_List', (Handle, 'h', InMode))
as_List_body = """
Handle h;
ListObject *l;
if (!PyArg_ParseTuple(_args, "O&", ResObj_Convert, &h))
        return NULL;
l = (ListObject *)ListObj_New(as_List(h));
l->ob_must_be_disposed = 0;
_res = Py_BuildValue("O", l);
return _res;
"""
f = ManualGenerator("as_List", as_List_body)
f.docstring = lambda: "(Resource)->List.\nReturns List object (which is not auto-freed!)"
functions.append(f)

f = Method(Handle, 'as_Resource', (ListHandle, 'lh', InMode))
methods.append(f)

# Manual generator for CreateCustomList, due to callback ideosyncracies
CreateCustomList_body = """\
Rect rView;
Rect dataBounds;
Point cellSize;

PyObject *listDefFunc;
ListDefSpec theSpec;
WindowPtr theWindow;
Boolean drawIt;
Boolean hasGrow;
Boolean scrollHoriz;
Boolean scrollVert;
ListHandle outList;

if (!PyArg_ParseTuple(_args, "O&O&O&(iO)O&bbbb",
                      PyMac_GetRect, &rView,
                      PyMac_GetRect, &dataBounds,
                      PyMac_GetPoint, &cellSize,
                      &theSpec.defType, &listDefFunc,
                      WinObj_Convert, &theWindow,
                      &drawIt,
                      &hasGrow,
                      &scrollHoriz,
                      &scrollVert))
        return NULL;


/* Carbon applications use the CreateCustomList API */
theSpec.u.userProc = myListDefFunctionUPP;
CreateCustomList(&rView,
                 &dataBounds,
                 cellSize,
                 &theSpec,
                 theWindow,
                 drawIt,
                 hasGrow,
                 scrollHoriz,
                 scrollVert,
                 &outList);


_res = ListObj_New(outList);
if (_res == NULL)
        return NULL;
Py_INCREF(listDefFunc);
((ListObject*)_res)->ob_ldef_func = listDefFunc;
return _res;\
"""

f = ManualGenerator("CreateCustomList", CreateCustomList_body);
f.docstring = lambda: "(Rect rView, Rect dataBounds, Point cellSize, ListDefSpec theSpec, WindowPtr theWindow, Boolean drawIt, Boolean hasGrow, Boolean scrollHoriz, Boolean scrollVert) -> (ListHandle outList)"
module.add(f)

# add the populated lists to the generator groups
# (in a different wordl the scan program would generate this)
for f in functions: module.add(f)
for f in methods: object.add(f)


# generate output (open the output file as late as possible)
SetOutputFileName(OUTPUTFILE)
module.generate()



import traceback
import sys
import W
import os
import types
from Carbon import List


class TraceBack:

    def __init__(self, title = "Traceback"):
        app = W.getapplication()  # checks if W is properly initialized
        self.title = title
        self.w = None
        self.closed = 1
        self.start = 0
        self.lastwindowtitle = ""
        self.bounds = (360, 298)

    def traceback(self, start = 0, lastwindowtitle = ""):
        try:
            self.lastwindowtitle = lastwindowtitle
            self.start = start
            self.type, self.value, self.tb = sys.exc_info()
            if self.type is not SyntaxError:
                self.show()
                if type(self.type) == types.ClassType:
                    errortext = self.type.__name__
                else:
                    errortext = str(self.type)
                value = str(self.value)
                if self.value and value:
                    errortext = errortext + ": " + value
                self.w.text.set(errortext)
                self.buildtblist()
                self.w.list.set(self.textlist)
                self.w.list.setselection([len(self.textlist) - 1])
                self.w.wid.SelectWindow()
                self.closed = 0
            else:
                self.syntaxerror()
        except:
            traceback.print_exc()

    def syntaxerror(self):
        try:
            value, (filename, lineno, charno, line) = self.value
        except:
            filename = ""
            lineno = None
            value = self.value
        if not filename and self.lastwindowtitle:
            filename = self.lastwindowtitle
        elif not filename:
            filename = "<unknown>"
        if filename and os.path.exists(filename):
            filename = os.path.split(filename)[1]
        if lineno and charno is not None:
            charno = charno - 1
            text = str(value) + '\rFile: "' + str(filename) + '", line ' + str(lineno) + '\r\r' + line[:charno] + "\xa5" + line[charno:-1]
        else:
            text = str(value) + '\rFile: "' + str(filename) + '"'
        self.syntaxdialog = W.ModalDialog((360, 120), "Syntax Error")
        self.syntaxdialog.text = W.TextBox((10, 10, -10, -40), text)
        self.syntaxdialog.cancel = W.Button((-190, -32, 80, 16), "Cancel", self.syntaxclose)
        self.syntaxdialog.edit = W.Button((-100, -32, 80, 16), "Edit", self.syntaxedit)
        self.syntaxdialog.setdefaultbutton(self.syntaxdialog.edit)
        self.syntaxdialog.bind("cmd.", self.syntaxdialog.cancel.push)
        self.syntaxdialog.open()

    def syntaxclose(self):
        self.syntaxdialog.close()
        del self.syntaxdialog

    def syntaxedit(self):
        try:
            value, (filename, lineno, charno, line) = self.value
        except:
            filename = ""
            lineno = None
        if not filename and self.lastwindowtitle:
            filename = self.lastwindowtitle
        elif not filename:
            filename = "<unknown>"
        self.syntaxclose()
        if lineno:
            if charno is None:
                charno = 1
            W.getapplication().openscript(filename, lineno, charno - 1)
        else:
            W.getapplication().openscript(filename)

    def show(self):
        if self.closed:
            self.setupwidgets()
            self.w.open()
        else:
            self.w.wid.ShowWindow()
            self.w.wid.SelectWindow()

    def hide(self):
        if self.closed:
            return
        self.w.close()

    def close(self):
        self.bounds = self.w.getbounds()
        self.closed = 1
        self.type, self.value, self.tb = None, None, None
        self.tblist = None

    def activate(self, onoff):
        if onoff:
            if self.closed:
                self.traceback()
            self.closed = 0
            self.checkbuttons()

    def setupwidgets(self):
        self.w = W.Window(self.bounds, self.title, minsize = (316, 168))
        self.w.text = W.TextBox((10, 10, -10, 30))
        self.w.tbtitle = W.TextBox((10, 40, -10, 10), "Traceback (innermost last):")
        self.w.list = W.TwoLineList((10, 60, -10, -40), callback = self.listhit)

        self.w.editbutton = W.Button((10, -30, 60, 16), "Edit", self.edit)
        self.w.editbutton.enable(0)

        self.w.browselocalsbutton = W.Button((80, -30, 100, 16), "Browse locals\xc9", self.browselocals)
        self.w.browselocalsbutton.enable(0)

        self.w.postmortembutton = W.Button((190, -30, 100, 16), "Post mortem\xc9", self.postmortem)

        self.w.setdefaultbutton(self.w.editbutton)
        self.w.bind("cmdb", self.w.browselocalsbutton.push)
        self.w.bind("<close>", self.close)
        self.w.bind("<activate>", self.activate)

    def buildtblist(self):
        tb = self.tb
        for i in range(self.start):
            if tb.tb_next is None:
                break
            tb = tb.tb_next
        self.tblist = traceback.extract_tb(tb)
        self.textlist = []
        for filename, lineno, func, line in self.tblist:
            tbline = ""
            if os.path.exists(filename):
                filename = os.path.split(filename)[1]
            tbline = 'File "%s", line %r, in %r' % (filename, lineno, func)
            if line:
                tbline = tbline + '\r      ' + line
            self.textlist.append(tbline[:255])

    def edit(self):
        sel = self.w.list.getselection()
        for i in sel:
            filename, lineno, func, line = self.tblist[i]
            W.getapplication().openscript(filename, lineno)

    def browselocals(self):
        sel = self.w.list.getselection()
        for i in sel:
            tb = self.tb
            for j in range(i + self.start):
                tb = tb.tb_next
            self.browse(tb.tb_frame.f_locals)

    def browse(self, object):
        import PyBrowser
        PyBrowser.Browser(object)

    def postmortem(self):
        import PyDebugger
        PyDebugger.postmortem(self.type, self.value, self.tb)

    def listhit(self, isdbl):
        if isdbl:
            self.w.editbutton.push()
        else:
            self.checkbuttons()

    def checkbuttons(self):
        havefile = len(self.w.list.getselection()) > 0
        self.w.editbutton.enable(havefile)
        self.w.browselocalsbutton.enable(havefile)
        self.w.setdefaultbutton(havefile and self.w.editbutton or self.w.postmortembutton)



# Scan an Apple header file, generating a Python file of generator calls.

import sys
import os
from bgenlocations import TOOLBOXDIR, BGENDIR
sys.path.append(BGENDIR)
from scantools import Scanner

WASTEDIR='/Users/jack/src/waste/C_C++ Headers/'

if not os.path.exists(WASTEDIR):
    raise 'Error: not found: %s', WASTEDIR

OBJECT = "TEHandle"
SHORT = "waste"
OBJECT = "WEReference"
OBJECT2 = "WEObjectReference"

def main():
    input = WASTEDIR + "WASTE.h"
    output = SHORT + "gen.py"
    defsoutput = os.path.join(os.path.split(TOOLBOXDIR)[0], "WASTEconst.py")
    scanner = MyScanner(input, output, defsoutput)
    scanner.scan()
##      scanner.gentypetest(SHORT+"typetest.py")
    scanner.close()
    print "=== Testing definitions output code ==="
    execfile(defsoutput, {}, {})
    print "=== Done scanning and generating, now importing the generated code... ==="
    exec "import " + SHORT + "support"
    print "=== Done.  It's up to you to compile it now! ==="

#class MyScanner(Scanner_PreUH3):
class MyScanner(Scanner):

    def destination(self, type, name, arglist):
        classname = "Function"
        listname = "functions"
        if arglist:
            t, n, m = arglist[-1]
            # This is non-functional today
            if t == OBJECT and m == "InMode":
                classname = "Method"
                listname = "methods"
            else:
                t, n, m = arglist[0]
                if t == OBJECT2 and m == "InMode":
                    classname = "Method2"
                    listname = "methods2"
        return classname, listname

    def writeinitialdefs(self):
        self.defsfile.write("kPascalStackBased = None # workaround for header parsing\n")
        self.defsfile.write("def FOUR_CHAR_CODE(x): return x\n")

    def makeblacklistnames(self):
        return [
                "WEDispose",
                "WESetInfo", # Argument type unknown...
                "WEGetInfo",
                "WEVersion", # Unfortunately...
                "WEPut", # XXXX TBD: needs array of flavortypes.
                "WEGetOneAttribute", # XXXX TBD: output buffer
                # Incompatible constant definitions
                "weDoAutoScroll",
                "weDoOutlineHilite",
                "weDoReadOnly",
                "weDoUndo",
                "weDoIntCutAndPaste",
                "weDoDragAndDrop",
                "weDoInhibitRecal",
                "weDoUseTempMem",
                "weDoDrawOffscreen",
                "weDoInhibitRedraw",
                "weDoMonoStyled",
                "weDoMultipleUndo",
                "weDoNoKeyboardSync",
                "weDoInhibitICSupport",
                "weDoInhibitColor",
                ]

    def makeblacklisttypes(self):
        return [
                "DragReference",        # For now...
                "UniversalProcPtr",
                "WEFontIDToNameUPP",
                "WEFontNameToIDUPP",
                "WEClickLoopUPP",
                "WEScrollUPP",
                "WETSMPreUpdateUPP",
                "WETSMPostUpdateUPP",
                "WEPreTrackDragUPP",
                "WETranslateDragUPP",
                "WEHiliteDropAreaUPP",
                "WEDrawTextUPP",
                "WEDrawTSMHiliteUPP",
                "WEPixelToCharUPP",
                "WECharToPixelUPP",
                "WELineBreakUPP",
                "WEWordBreakUPP",
                "WECharByteUPP",
                "WECharTypeUPP",
                "WEEraseUPP",
                "WEFluxUPP",
                "WENewObjectUPP",
                "WEDisposeObjectUPP",
                "WEDrawObjectUPP",
                "WEClickObjectUPP",
                "WEStreamObjectUPP",
                "WEHoverObjectUPP",
                "WERuler",              # XXXX To be done
                "WERuler_ptr",  # ditto
                "WEParaInfo",   # XXXX To be done
                "WEPrintSession",       # XXXX To be done
                "WEPrintOptions_ptr", # XXXX To be done
                ]

    def makerepairinstructions(self):
        return [
                ([("void_ptr", "*", "InMode"), ("SInt32", "*", "InMode")],
                 [("InBuffer", "*", "*")]),

                # WEContinuousStyle
                ([("WEStyleMode", "ioMode", "OutMode"), ("TextStyle", "outTextStyle", "OutMode")],
                 [("WEStyleMode", "*", "InOutMode"), ("TextStyle", "*", "*")]),

                # WECopyRange
                ([('Handle', 'outText', 'InMode'), ('StScrpHandle', 'outStyles', 'InMode'),
                ('WESoupHandle', 'outSoup', 'InMode')],
         [('OptHandle', '*', '*'), ('OptStScrpHandle', '*', '*'),
                ('OptSoupHandle', '*', '*')]),

                # WEInsert
                ([('StScrpHandle', 'inStyles', 'InMode'), ('WESoupHandle', 'inSoup', 'InMode')],
         [('OptStScrpHandle', '*', '*'), ('OptSoupHandle', '*', '*')]),

        # WEGetObjectOwner
        ("WEGetObjectOwner",
         [('WEReference', '*', 'ReturnMode')],
         [('ExistingWEReference', '*', 'ReturnMode')]),

        # WEFindParagraph
        ([("char_ptr", "inKey", "InMode")],
         [("stringptr", "*", "*")]),

                # WESetOneAttribute
                ([("void_ptr", "*", "InMode"), ("ByteCount", "*", "InMode")],
                 [("InBuffer", "*", "*")]),
                ]

if __name__ == "__main__":
    main()



import os
from Carbon import Qd
from Carbon import Win
from Carbon import Qt, QuickTime
import W
from Carbon import File
from Carbon import Evt, Events

_moviesinitialized = 0

def EnterMovies():
    global _moviesinitialized
    if not _moviesinitialized:
        Qt.EnterMovies()
        _moviesinitialized = 1

class Movie(W.Widget):

    def __init__(self, possize):
        EnterMovies()
        self.movie = None
        self.running = 0
        W.Widget.__init__(self, possize)

    def adjust(self, oldbounds):
        self.SetPort()
        self.GetWindow().InvalWindowRect(oldbounds)
        self.GetWindow().InvalWindowRect(self._bounds)
        self.calcmoviebox()

    def set(self, path_or_fss, start = 0):
        self.SetPort()
        if self.movie:
            #self.GetWindow().InvalWindowRect(self.movie.GetMovieBox())
            Qd.PaintRect(self.movie.GetMovieBox())
        path = File.pathname(path)
        self.movietitle = os.path.basename(path)
        movieResRef = Qt.OpenMovieFile(path_or_fss, 1)
        self.movie, dummy, dummy = Qt.NewMovieFromFile(movieResRef, 0, QuickTime.newMovieActive)
        self.moviebox = self.movie.GetMovieBox()
        self.calcmoviebox()
        Qd.ObscureCursor()      # XXX does this work at all?
        self.movie.GoToBeginningOfMovie()
        if start:
            self.movie.StartMovie()
            self.running = 1
        else:
            self.running = 0
            self.movie.MoviesTask(0)

    def get(self):
        return self.movie

    def getmovietitle(self):
        return self.movietitle

    def start(self):
        if self.movie:
            Qd.ObscureCursor()
            self.movie.StartMovie()
            self.running = 1

    def stop(self):
        if self.movie:
            self.movie.StopMovie()
            self.running = 0

    def rewind(self):
        if self.movie:
            self.movie.GoToBeginningOfMovie()

    def calcmoviebox(self):
        if not self.movie:
            return
        ml, mt, mr, mb = self.moviebox
        wl, wt, wr, wb = widgetbox = self._bounds
        mheight = mb - mt
        mwidth = mr - ml
        wheight = wb - wt
        wwidth = wr - wl
        if (mheight * 2 < wheight) and (mwidth * 2 < wwidth):
            scale = 2
        elif mheight > wheight or mwidth > wwidth:
            scale = min(float(wheight) / mheight, float(wwidth) / mwidth)
        else:
            scale = 1
        mwidth, mheight = mwidth * scale, mheight * scale
        ml, mt = wl + (wwidth - mwidth) / 2, wt + (wheight - mheight) / 2
        mr, mb = ml + mwidth, mt + mheight
        self.movie.SetMovieBox((ml, mt, mr, mb))

    def idle(self, *args):
        if self.movie:
            if not self.movie.IsMovieDone() and self.running:
                Qd.ObscureCursor()
                while 1:
                    self.movie.MoviesTask(0)
                    gotone, event = Evt.EventAvail(Events.everyEvent)
                    if gotone or self.movie.IsMovieDone():
                        break
            elif self.running:
                box = self.movie.GetMovieBox()
                self.SetPort()
                self.GetWindow().InvalWindowRect(box)
                self.movie = None
                self.running = 0

    def draw(self, visRgn = None):
        if self._visible:
            Qd.PaintRect(self._bounds)
            if self.movie:
                self.movie.UpdateMovie()
                self.movie.MoviesTask(0)



from Carbon import App, Evt, Qd, QuickDraw, Win
import string
from types import *
import sys

class WidgetsError(Exception): pass

DEBUG = 0


def _intRect((l, t, r, b)):
    return (int(l), int(t), int(r), int(b))


class Widget:

    """Base class for all widgets."""

    _selectable = 0

    def __init__(self, possize):
        self._widgets = []
        self._widgetsdict = {}
        self._possize = possize
        self._bounds = None
        self._visible = 1
        self._enabled = 0
        self._selected = 0
        self._activated = 0
        self._callback = None
        self._parent = None
        self._parentwindow = None
        self._bindings = {}
        self._backcolor = None

    def show(self, onoff):
        self._visible = onoff
        for w in self._widgets:
            w.show(onoff)
        if self._parentwindow is not None and self._parentwindow.wid is not None:
            self.SetPort()
            if onoff:
                self.draw()
            else:
                Qd.EraseRect(self._bounds)

    def draw(self, visRgn = None):
        if self._visible:
            # draw your stuff here
            pass

    def getpossize(self):
        return self._possize

    def getbounds(self):
        return self._bounds

    def move(self, x, y = None):
        """absolute move"""
        if y == None:
            x, y = x
        if type(self._possize) <> TupleType:
            raise WidgetsError, "can't move widget with bounds function"
        l, t, r, b = self._possize
        self.resize(x, y, r, b)

    def rmove(self, x, y = None):
        """relative move"""
        if y == None:
            x, y = x
        if type(self._possize) <> TupleType:
            raise WidgetsError, "can't move widget with bounds function"
        l, t, r, b = self._possize
        self.resize(l + x, t + y, r, b)

    def resize(self, *args):
        if len(args) == 1:
            if type(args[0]) == FunctionType or type(args[0]) == MethodType:
                self._possize = args[0]
            else:
                apply(self.resize, args[0])
        elif len(args) == 2:
            self._possize = (0, 0) + args
        elif len(args) == 4:
            self._possize = args
        else:
            raise TypeError, "wrong number of arguments"
        self._calcbounds()

    def open(self):
        self._calcbounds()

    def close(self):
        del self._callback
        del self._possize
        del self._bindings
        del self._parent
        del self._parentwindow

    def bind(self, key, callback):
        """bind a key or an 'event' to a callback"""
        if callback:
            self._bindings[key] = callback
        elif self._bindings.has_key(key):
            del self._bindings[key]

    def adjust(self, oldbounds):
        self.SetPort()
        self.GetWindow().InvalWindowRect(oldbounds)
        self.GetWindow().InvalWindowRect(self._bounds)

    def _calcbounds(self):
        # calculate absolute bounds relative to the window origin from our
        # abstract _possize attribute, which is either a 4-tuple or a callable object
        oldbounds = self._bounds
        pl, pt, pr, pb = self._parent._bounds
        if callable(self._possize):
            # _possize is callable, let it figure it out by itself: it should return
            # the bounds relative to our parent widget.
            width = pr - pl
            height = pb - pt
            self._bounds = Qd.OffsetRect(_intRect(self._possize(width, height)), pl, pt)
        else:
            # _possize must be a 4-tuple. This is where the algorithm by Peter Kriens and
            # Petr van Blokland kicks in. (*** Parts of this algorithm are applied for
            # patents by Ericsson, Sweden ***)
            l, t, r, b = self._possize
            # depending on the values of l(eft), t(op), r(right) and b(ottom),
            # they mean different things:
            if l < -1:
                # l is less than -1, this mean it measures from the *right* of it's parent
                l = pr + l
            else:
                # l is -1 or greater, this mean it measures from the *left* of it's parent
                l = pl + l
            if t < -1:
                # t is less than -1, this mean it measures from the *bottom* of it's parent
                t = pb + t
            else:
                # t is -1 or greater, this mean it measures from the *top* of it's parent
                t = pt + t
            if r > 1:
                # r is greater than 1, this means r is the *width* of the widget
                r = l + r
            else:
                # r is less than 1, this means it measures from the *right* of it's parent
                r = pr + r
            if b > 1:
                # b is greater than 1, this means b is the *height* of the widget
                b = t + b
            else:
                # b is less than 1, this means it measures from the *bottom* of it's parent
                b = pb + b
            self._bounds = (l, t, r, b)
        if oldbounds and oldbounds <> self._bounds:
            self.adjust(oldbounds)
        for w in self._widgets:
            w._calcbounds()

    def test(self, point):
        if Qd.PtInRect(point, self._bounds):
            return 1

    def click(self, point, modifiers):
        pass

    def findwidget(self, point, onlyenabled = 1):
        if self.test(point):
            for w in self._widgets:
                widget = w.findwidget(point)
                if widget is not None:
                    return widget
            if self._enabled or not onlyenabled:
                return self

    def forall(self, methodname, *args):
        for w in self._widgets:
            rv = apply(w.forall, (methodname,) + args)
            if rv:
                return rv
        if self._bindings.has_key("<" + methodname + ">"):
            callback = self._bindings["<" + methodname + ">"]
            rv = apply(callback, args)
            if rv:
                return rv
        if hasattr(self, methodname):
            method = getattr(self, methodname)
            return apply(method, args)

    def forall_butself(self, methodname, *args):
        for w in self._widgets:
            rv = apply(w.forall, (methodname,) + args)
            if rv:
                return rv

    def forall_frombottom(self, methodname, *args):
        if self._bindings.has_key("<" + methodname + ">"):
            callback = self._bindings["<" + methodname + ">"]
            rv = apply(callback, args)
            if rv:
                return rv
        if hasattr(self, methodname):
            method = getattr(self, methodname)
            rv = apply(method, args)
            if rv:
                return rv
        for w in self._widgets:
            rv = apply(w.forall_frombottom, (methodname,) + args)
            if rv:
                return rv

    def _addwidget(self, key, widget):
        if widget in self._widgets:
            raise ValueError, "duplicate widget"
        if self._widgetsdict.has_key(key):
            self._removewidget(key)
        self._widgets.append(widget)
        self._widgetsdict[key] = widget
        widget._parent = self
        self._setparentwindow(widget)
        if self._parentwindow and self._parentwindow.wid:
            widget.forall_frombottom("open")
            self.GetWindow().InvalWindowRect(widget._bounds)

    def _setparentwindow(self, widget):
        widget._parentwindow = self._parentwindow
        for w in widget._widgets:
            self._setparentwindow(w)

    def _removewidget(self, key):
        if not self._widgetsdict.has_key(key):
            raise KeyError, "no widget with key %r" % (key,)
        widget = self._widgetsdict[key]
        for k in widget._widgetsdict.keys():
            widget._removewidget(k)
        if self._parentwindow._currentwidget == widget:
            widget.select(0)
            self._parentwindow._currentwidget = None
        self.SetPort()
        self.GetWindow().InvalWindowRect(widget._bounds)
        widget.close()
        del self._widgetsdict[key]
        self._widgets.remove(widget)

    def __setattr__(self, attr, value):
        if type(value) == InstanceType and isinstance(value, Widget) and        \
                        attr not in ("_currentwidget", "_lastrollover",
                        "_parent", "_parentwindow", "_defaultbutton"):
            if hasattr(self, attr):
                raise ValueError, "Can't replace existing attribute: " + attr
            self._addwidget(attr, value)
        self.__dict__[attr] = value

    def __delattr__(self, attr):
        if attr == "_widgetsdict":
            raise AttributeError, "cannot delete attribute _widgetsdict"
        if self._widgetsdict.has_key(attr):
            self._removewidget(attr)
            if self.__dict__.has_key(attr):
                del self.__dict__[attr]
        elif self.__dict__.has_key(attr):
            del self.__dict__[attr]
        else:
            raise AttributeError, attr

    def __setitem__(self, key, value):
        self._addwidget(key, value)

    def __getitem__(self, key):
        if not self._widgetsdict.has_key(key):
            raise KeyError, key
        return self._widgetsdict[key]

    def __delitem__(self, key):
        self._removewidget(key)

    def SetPort(self):
        self._parentwindow.SetPort()


    def GetWindow(self):
        return self._parentwindow.GetWindow()

    def __del__(self):
        if DEBUG:
            print "%s instance deleted" % self.__class__.__name__

    def _drawbounds(self):
        Qd.FrameRect(self._bounds)


class ClickableWidget(Widget):

    """Base class for clickable widgets. (note: self._enabled must be true to receive click events.)"""

    def click(self, point, modifiers):
        pass

    def enable(self, onoff):
        self._enabled = onoff
        self.SetPort()
        self.draw()

    def callback(self):
        if self._callback:
            return CallbackCall(self._callback, 1)


class SelectableWidget(ClickableWidget):

    """Base class for selectable widgets."""

    _selectable = 1

    def select(self, onoff, isclick = 0):
        if onoff == self._selected:
            return 1
        if self._bindings.has_key("<select>"):
            callback = self._bindings["<select>"]
            if callback(onoff):
                return 1
        self._selected = onoff
        if onoff:
            if self._parentwindow._currentwidget is not None:
                self._parentwindow._currentwidget.select(0)
            self._parentwindow._currentwidget = self
        else:
            self._parentwindow._currentwidget = None

    def key(self, char, event):
        pass

    def drawselframe(self, onoff):
        if not self._parentwindow._hasselframes:
            return
        App.DrawThemeFocusRect(self._bounds, onoff)

    def adjust(self, oldbounds):
        self.SetPort()
        if self._selected:
            self.GetWindow().InvalWindowRect(Qd.InsetRect(oldbounds, -3, -3))
            self.GetWindow().InvalWindowRect(Qd.InsetRect(self._bounds, -3, -3))
        else:
            self.GetWindow().InvalWindowRect(oldbounds)
            self.GetWindow().InvalWindowRect(self._bounds)


class _Line(Widget):

    def __init__(self, possize, thickness = 1):
        Widget.__init__(self, possize)
        self._thickness = thickness

    def open(self):
        self._calcbounds()
        self.SetPort()
        self.draw()

    def draw(self, visRgn = None):
        if self._visible:
            Qd.PaintRect(self._bounds)

    def _drawbounds(self):
        pass

class HorizontalLine(_Line):

    def _calcbounds(self):
        Widget._calcbounds(self)
        l, t, r, b = self._bounds
        self._bounds = l, t, r, t + self._thickness

class VerticalLine(_Line):

    def _calcbounds(self):
        Widget._calcbounds(self)
        l, t, r, b = self._bounds
        self._bounds = l, t, l + self._thickness, b


class Frame(Widget):

    def __init__(self, possize, pattern = Qd.GetQDGlobalsBlack(), color = (0, 0, 0)):
        Widget.__init__(self, possize)
        self._framepattern = pattern
        self._framecolor = color

    def setcolor(self, color):
        self._framecolor = color
        self.SetPort()
        self.draw()

    def setpattern(self, pattern):
        self._framepattern = pattern
        self.SetPort()
        self.draw()

    def draw(self, visRgn = None):
        if self._visible:
            penstate = Qd.GetPenState()
            Qd.PenPat(self._framepattern)
            Qd.RGBForeColor(self._framecolor)
            Qd.FrameRect(self._bounds)
            Qd.RGBForeColor((0, 0, 0))
            Qd.SetPenState(penstate)

def _darkencolor((r, g, b)):
    return int(0.75 * r), int(0.75 * g), int(0.75 * b)

class BevelBox(Widget):

    """'Platinum' beveled rectangle."""

    def __init__(self, possize, color = (0xe000, 0xe000, 0xe000)):
        Widget.__init__(self, possize)
        self._color = color
        self._darkercolor = _darkencolor(color)

    def setcolor(self, color):
        self._color = color
        self.SetPort()
        self.draw()

    def draw(self, visRgn = None):
        if self._visible:
            l, t, r, b = Qd.InsetRect(self._bounds, 1, 1)
            Qd.RGBForeColor(self._color)
            Qd.PaintRect((l, t, r, b))
            Qd.RGBForeColor(self._darkercolor)
            Qd.MoveTo(l, b)
            Qd.LineTo(r, b)
            Qd.LineTo(r, t)
            Qd.RGBForeColor((0, 0, 0))


class Group(Widget):

    """A container for subwidgets"""


class HorizontalPanes(Widget):

    """Panes, a.k.a. frames. Works a bit like a group. Devides the widget area into "panes",
    which can be resized by the user by clicking and dragging between the subwidgets."""

    _direction = 1

    def __init__(self, possize, panesizes = None, gutter = 8):
        """panesizes should be a tuple of numbers. The length of the tuple is the number of panes,
        the items in the tuple are the relative sizes of these panes; these numbers should add up
        to 1 (the total size of all panes)."""
        Widget.__init__(self, possize)
        self._panesizes = panesizes
        self._gutter = gutter
        self._enabled = 1
        self.setuppanes()

    #def open(self):
    #       self.installbounds()
    #       ClickableWidget.open(self)

    def _calcbounds(self):
        # hmmm. It should not neccesary be override _calcbounds :-(
        self.installbounds()
        Widget._calcbounds(self)

    def setuppanes(self):
        panesizes = self._panesizes
        total = 0
        if panesizes is not None:
            #if len(self._widgets) <> len(panesizes):
            #       raise TypeError, 'number of widgets does not match number of panes'
            for panesize in panesizes:
                if not 0 < panesize < 1:
                    raise TypeError, 'pane sizes must be between 0 and 1, not including.'
                total = total + panesize
            if round(total, 4) <> 1.0:
                raise TypeError, 'pane sizes must add up to 1'
        else:
            # XXX does not work!
            step = 1.0 / len(self._widgets)
            panesizes = []
            for i in range(len(self._widgets)):
                panesizes.append(step)
        current = 0
        self._panesizes = []
        self._gutters = []
        for panesize in panesizes:
            if current:
                self._gutters.append(current)
            self._panesizes.append((current, current + panesize))
            current = current + panesize
        self.makepanebounds()

    def getpanesizes(self):
        return map(lambda (fr, to): to-fr,  self._panesizes)

    boundstemplate = "lambda width, height: (0, height * %s + %d, width, height * %s + %d)"

    def makepanebounds(self):
        halfgutter = self._gutter / 2
        self._panebounds = []
        for i in range(len(self._panesizes)):
            panestart, paneend = self._panesizes[i]
            boundsstring = self.boundstemplate % (repr(panestart), panestart and halfgutter,
                                            repr(paneend), (paneend <> 1.0) and -halfgutter)
            self._panebounds.append(eval(boundsstring))

    def installbounds(self):
        #self.setuppanes()
        for i in range(len(self._widgets)):
            w = self._widgets[i]
            w._possize = self._panebounds[i]
            #if hasattr(w, "setuppanes"):
            #       w.setuppanes()
            if hasattr(w, "installbounds"):
                w.installbounds()

    def rollover(self, point, onoff):
        if onoff:
            orgmouse = point[self._direction]
            halfgutter = self._gutter / 2
            l, t, r, b = self._bounds
            if self._direction:
                begin, end = t, b
            else:
                begin, end = l, r

            i = self.findgutter(orgmouse, begin, end)
            if i is None:
                SetCursor("arrow")
            else:
                SetCursor(self._direction and 'vmover' or 'hmover')

    def findgutter(self, orgmouse, begin, end):
        tolerance = max(4, self._gutter) / 2
        for i in range(len(self._gutters)):
            pos = begin + (end - begin) * self._gutters[i]
            if abs(orgmouse - pos) <= tolerance:
                break
        else:
            return
        return i

    def click(self, point, modifiers):
        # what a mess...
        orgmouse = point[self._direction]
        halfgutter = self._gutter / 2
        l, t, r, b = self._bounds
        if self._direction:
            begin, end = t, b
        else:
            begin, end = l, r

        i = self.findgutter(orgmouse, begin, end)
        if i is None:
            return

        pos = orgpos = begin + (end - begin) * self._gutters[i] # init pos too, for fast click on border, bug done by Petr

        minpos = self._panesizes[i][0]
        maxpos = self._panesizes[i+1][1]
        minpos = begin + (end - begin) * minpos + 64
        maxpos = begin + (end - begin) * maxpos - 64
        if minpos > orgpos and maxpos < orgpos:
            return

        #SetCursor("fist")
        self.SetPort()
        if self._direction:
            rect = l, orgpos - 1, r, orgpos
        else:
            rect = orgpos - 1, t, orgpos, b

        # track mouse --- XXX  move to separate method?
        Qd.PenMode(QuickDraw.srcXor)
        Qd.PenPat(Qd.GetQDGlobalsGray())
        Qd.PaintRect(_intRect(rect))
        lastpos = None
        while Evt.Button():
            pos = orgpos - orgmouse + Evt.GetMouse()[self._direction]
            pos = max(pos, minpos)
            pos = min(pos, maxpos)
            if pos == lastpos:
                continue
            Qd.PenPat(Qd.GetQDGlobalsGray())
            Qd.PaintRect(_intRect(rect))
            if self._direction:
                rect = l, pos - 1, r, pos
            else:
                rect = pos - 1, t, pos, b
            Qd.PenPat(Qd.GetQDGlobalsGray())
            Qd.PaintRect(_intRect(rect))
            lastpos = pos
            self._parentwindow.wid.GetWindowPort().QDFlushPortBuffer(None)
            Evt.WaitNextEvent(0, 3)
        Qd.PaintRect(_intRect(rect))
        Qd.PenNormal()
        SetCursor("watch")

        newpos = (pos - begin) / float(end - begin)
        self._gutters[i] = newpos
        self._panesizes[i] = self._panesizes[i][0], newpos
        self._panesizes[i+1] = newpos, self._panesizes[i+1][1]
        self.makepanebounds()
        self.installbounds()
        self._calcbounds()


class VerticalPanes(HorizontalPanes):
    """see HorizontalPanes"""
    _direction = 0
    boundstemplate = "lambda width, height: (width * %s + %d, 0, width * %s + %d, height)"


class ColorPicker(ClickableWidget):

    """Color picker widget. Allows the user to choose a color."""

    def __init__(self, possize, color = (0, 0, 0), callback = None):
        ClickableWidget.__init__(self, possize)
        self._color = color
        self._callback = callback
        self._enabled = 1

    def click(self, point, modifiers):
        if not self._enabled:
            return
        import ColorPicker
        newcolor, ok = ColorPicker.GetColor("", self._color)
        if ok:
            self._color = newcolor
            self.SetPort()
            self.draw()
            if self._callback:
                return CallbackCall(self._callback, 0, self._color)

    def set(self, color):
        self._color = color
        self.SetPort()
        self.draw()

    def get(self):
        return self._color

    def draw(self, visRgn=None):
        if self._visible:
            if not visRgn:
                visRgn = self._parentwindow.wid.GetWindowPort().visRgn
            Qd.PenPat(Qd.GetQDGlobalsGray())
            rect = self._bounds
            Qd.FrameRect(rect)
            rect = Qd.InsetRect(rect, 3, 3)
            Qd.PenNormal()
            Qd.RGBForeColor(self._color)
            Qd.PaintRect(rect)
            Qd.RGBForeColor((0, 0, 0))


# misc utils

def CallbackCall(callback, mustfit, *args):
    """internal helper routine for W"""
    # XXX this function should die.
    if type(callback) == FunctionType:
        func = callback
        maxargs = func.func_code.co_argcount
    elif type(callback) == MethodType:
        func = callback.im_func
        maxargs = func.func_code.co_argcount - 1
    else:
        if callable(callback):
            return apply(callback, args)
        else:
            raise TypeError, "uncallable callback object"

    if func.func_defaults:
        minargs = maxargs - len(func.func_defaults)
    else:
        minargs = maxargs
    if minargs <= len(args) <= maxargs:
        return apply(callback, args)
    elif not mustfit and minargs == 0:
        return callback()
    else:
        if mustfit:
            raise TypeError, "callback accepts wrong number of arguments: %r" % len(args)
        else:
            raise TypeError, "callback accepts wrong number of arguments: 0 or %r" % len(args)


def HasBaseClass(obj, class_):
    try:
        raise obj
    except class_:
        return 1
    except:
        pass
    return 0


#
# To remove the dependence of Widgets.rsrc we hardcode the cursor
# data below.
#_cursors = {
#       "watch" : Qd.GetCursor(QuickDraw.watchCursor).data,
#       "arrow" : Qd.GetQDGlobalsArrow(),
#       "iBeam" : Qd.GetCursor(QuickDraw.iBeamCursor).data,
#       "cross" : Qd.GetCursor(QuickDraw.crossCursor).data,
#       "plus"          : Qd.GetCursor(QuickDraw.plusCursor).data,
#       "hand"  : Qd.GetCursor(468).data,
#       "fist"          : Qd.GetCursor(469).data,
#       "hmover"        : Qd.GetCursor(470).data,
#       "vmover"        : Qd.GetCursor(471).data,
#       "zoomin"        : Qd.GetCursor(472).data,
#       "zoomout"       : Qd.GetCursor(473).data,
#       "zoom"  : Qd.GetCursor(474).data,
#}

_cursors = {
        'arrow':
        '\x00\x00\x40\x00\x60\x00\x70\x00\x78\x00\x7c\x00\x7e\x00\x7f\x00'
        '\x7f\x80\x7c\x00\x6c\x00\x46\x00\x06\x00\x03\x00\x03\x00\x00\x00'
        '\xc0\x00\xe0\x00\xf0\x00\xf8\x00\xfc\x00\xfe\x00\xff\x00\xff\x80'
        '\xff\xc0\xff\xe0\xfe\x00\xef\x00\xcf\x00\x87\x80\x07\x80\x03\x80'
        '\x00\x01\x00\x01',
        'cross':
        '\x04\x00\x04\x00\x04\x00\x04\x00\x04\x00\xff\xe0\x04\x00\x04\x00'
        '\x04\x00\x04\x00\x04\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        '\x00\x05\x00\x05',
        'fist':
        '\x00\x00\x00\x00\x0d\x80\x12\x70\x12\x4c\x12\x4a\x28\x0a\x28\x02'
        '\x48\x02\x40\x02\x20\x02\x20\x04\x10\x04\x08\x08\x04\x08\x04\x08'
        '\x00\x00\x00\x00\x0d\x80\x1f\xf0\x1f\xfc\x1f\xfe\x3f\xfe\x3f\xfe'
        '\x7f\xfe\x7f\xfe\x3f\xfe\x3f\xfc\x1f\xfc\x0f\xf8\x07\xf8\x07\xf8'
        '\x00\x09\x00\x08',
        'hand':
        '\x01\x80\x1a\x70\x26\x48\x26\x4a\x12\x4d\x12\x49\x68\x09\x98\x01'
        '\x88\x02\x40\x02\x20\x02\x20\x04\x10\x04\x08\x08\x04\x08\x04\x08'
        '\x01\x80\x1b\xf0\x3f\xf8\x3f\xfa\x1f\xff\x1f\xff\x6f\xff\xff\xff'
        '\xff\xfe\x7f\xfe\x3f\xfe\x3f\xfc\x1f\xfc\x0f\xf8\x07\xf8\x07\xf8'
        '\x00\x09\x00\x08',
        'hmover':
        '\x00\x00\x01\x80\x01\x80\x01\x80\x01\x80\x11\x88\x31\x8c\x7f\xfe'
        '\x31\x8c\x11\x88\x01\x80\x01\x80\x01\x80\x01\x80\x00\x00\x00\x00'
        '\x03\xc0\x03\xc0\x03\xc0\x03\xc0\x1b\xd8\x3b\xdc\x7f\xfe\xff\xff'
        '\x7f\xfe\x3b\xdc\x1b\xd8\x03\xc0\x03\xc0\x03\xc0\x03\xc0\x00\x00'
        '\x00\x07\x00\x07',
        'iBeam':
        '\x0c\x60\x02\x80\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00'
        '\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00\x02\x80\x0c\x60'
        '\x0c\x60\x02\x80\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00'
        '\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00\x01\x00\x02\x80\x0c\x60'
        '\x00\x04\x00\x07',
        'plus':
        '\x00\x00\x07\xc0\x04\x60\x04\x60\x04\x60\x7c\x7c\x43\x86\x42\x86'
        '\x43\x86\x7c\x7e\x3c\x7e\x04\x60\x04\x60\x07\xe0\x03\xe0\x00\x00'
        '\x0f\xc0\x0f\xe0\x0f\xf0\x0f\xf0\xff\xff\xff\xfe\xfc\x7f\xfc\x7f'
        '\xfc\x7f\xff\xff\x7f\xff\x7f\xff\x0f\xf0\x0f\xf0\x07\xf0\x03\xe0'
        '\x00\x08\x00\x08',
        'vmover':
        '\x00\x00\x01\x00\x03\x80\x07\xc0\x01\x00\x01\x00\x01\x00\x7f\xfc'
        '\x7f\xfc\x01\x00\x01\x00\x01\x00\x07\xc0\x03\x80\x01\x00\x00\x00'
        '\x01\x00\x03\x80\x07\xc0\x0f\xe0\x0f\xe0\x03\x80\xff\xfe\xff\xfe'
        '\xff\xfe\xff\xfe\x03\x80\x0f\xe0\x0f\xe0\x07\xc0\x03\x80\x01\x00'
        '\x00\x07\x00\x07',
        'watch':
        '\x3f\x00\x3f\x00\x3f\x00\x3f\x00\x40\x80\x84\x40\x84\x40\x84\x60'
        '\x9c\x60\x80\x40\x80\x40\x40\x80\x3f\x00\x3f\x00\x3f\x00\x3f\x00'
        '\x3f\x00\x3f\x00\x3f\x00\x3f\x00\x7f\x80\xff\xc0\xff\xc0\xff\xc0'
        '\xff\xc0\xff\xc0\xff\xc0\x7f\x80\x3f\x00\x3f\x00\x3f\x00\x3f\x00'
        '\x00\x08\x00\x08',
        'zoom':
        '\x0f\x00\x30\xc0\x40\x20\x40\x20\x80\x10\x80\x10\x80\x10\x80\x10'
        '\x40\x20\x40\x20\x30\xf0\x0f\x38\x00\x1c\x00\x0e\x00\x07\x00\x02'
        '\x0f\x00\x3f\xc0\x7f\xe0\x7f\xe0\xff\xf0\xff\xf0\xff\xf0\xff\xf0'
        '\x7f\xe0\x7f\xe0\x3f\xf0\x0f\x38\x00\x1c\x00\x0e\x00\x07\x00\x02'
        '\x00\x06\x00\x06',
        'zoomin':
        '\x0f\x00\x30\xc0\x40\x20\x46\x20\x86\x10\x9f\x90\x9f\x90\x86\x10'
        '\x46\x20\x40\x20\x30\xf0\x0f\x38\x00\x1c\x00\x0e\x00\x07\x00\x02'
        '\x0f\x00\x3f\xc0\x7f\xe0\x7f\xe0\xff\xf0\xff\xf0\xff\xf0\xff\xf0'
        '\x7f\xe0\x7f\xe0\x3f\xf0\x0f\x38\x00\x1c\x00\x0e\x00\x07\x00\x02'
        '\x00\x06\x00\x06',
        'zoomout':
        '\x0f\x00\x30\xc0\x40\x20\x40\x20\x80\x10\x9f\x90\x9f\x90\x80\x10'
        '\x40\x20\x40\x20\x30\xf0\x0f\x38\x00\x1c\x00\x0e\x00\x07\x00\x02'
        '\x0f\x00\x3f\xc0\x7f\xe0\x7f\xe0\xff\xf0\xff\xf0\xff\xf0\xff\xf0'
        '\x7f\xe0\x7f\xe0\x3f\xf0\x0f\x38\x00\x1c\x00\x0e\x00\x07\x00\x02'
        '\x00\x06\x00\x06',
}

def SetCursor(what):
    """Set the cursorshape to any of these: 'arrow', 'cross', 'fist', 'hand', 'hmover', 'iBeam',
    'plus', 'vmover', 'watch', 'zoom', 'zoomin', 'zoomout'."""
    Qd.SetCursor(_cursors[what])



# Scan an Apple header file, generating a Python file of generator calls.

import sys
import os
from bgenlocations import TOOLBOXDIR, BGENDIR
sys.path.append(BGENDIR)
from scantools import Scanner

LONG = "Fonts"
SHORT = "fm"

def main():
    input = "Fonts.h"
    output = SHORT + "gen.py"
    defsoutput = TOOLBOXDIR + LONG + ".py"
    scanner = MyScanner(input, output, defsoutput)
    scanner.scan()
    scanner.close()
    print "=== Testing definitions output code ==="
    execfile(defsoutput, {}, {})
    print "=== Done scanning and generating, now importing the generated code... ==="
    exec "import " + SHORT + "support"
    print "=== Done.  It's up to you to compile it now! ==="

class MyScanner(Scanner):

    def destination(self, type, name, arglist):
        classname = "Function"
        listname = "functions"
        return classname, listname

    def makeblacklistnames(self):
        return [
                "OutlineMetrics",       # Too complicated
                "AntiTextIsAntiAliased",        # XXXX Missing from library...
                "AntiTextGetEnabled",
                "AntiTextSetEnabled",
                "AntiTextGetApplicationAware",
                "AntiTextSetApplicationAware",
                # These are tricky: they're not Carbon dependent or anything, but they
                # exist only on 8.6 or later (both in Carbon and Classic).
                # Disabling them is the easiest path.
                'SetAntiAliasedTextEnabled',
                'IsAntiAliasedTextEnabled',
                # OS8-only
                'InitFonts',
                'SetFontLock',
                'FlushFonts',
                ]

    def makeblacklisttypes(self):
        return [
                "FMInput_ptr",  # Not needed for now
                "FMOutPtr",             # Ditto
##                      "void_ptr",             # Don't know how to do this right now
                "FontInfo",             # Ditto
                ]

    def makerepairinstructions(self):
        return [
                ([('Str255', '*', 'InMode')], [('Str255', '*', 'OutMode')]),
                ([('FMetricRecPtr', 'theMetrics', 'InMode')], [('FMetricRecPtr', 'theMetrics', 'OutMode')]),
                ([('short', 'byteCount', 'InMode'), ('void_ptr', 'textAddr', 'InMode'),],
                 [('TextBuffer', 'inText', 'InMode')]),
                ]

    def writeinitialdefs(self):
        self.defsfile.write("def FOUR_CHAR_CODE(x): return x\n")
        self.defsfile.write("kNilOptions = 0\n")

if __name__ == "__main__":
    main()



"""distutils

The main package for the Python Module Distribution Utilities.  Normally
used from a setup script as

   from distutils.core import setup

   setup (...)
"""

# This module should be kept compatible with Python 1.5.2.

__revision__ = "$Id: __init__.py,v 1.25 2004/07/23 19:47:32 theller Exp $"

__version__ = "2.4.0"



import sys
import bdb
import types
import os

import W
import WASTEconst
import PyBrowser
from Carbon import Qd
from Carbon import Evt
from Carbon import Lists
import MacOS
_filenames = {}

SIMPLE_TYPES = (
        types.NoneType,
        types.IntType,
        types.LongType,
        types.FloatType,
        types.ComplexType,
        types.StringType
)


class Debugger(bdb.Bdb):

    def __init__(self, title = 'Debugger'):
        bdb.Bdb.__init__(self)
        self.closed = 1
        self.title = title
        self.breaksviewer = None
        self.reset()
        self.tracing = 0
        self.tracingmonitortime = Evt.TickCount()
        self.editors = {}

        prefs = W.getapplication().getprefs()
        if prefs.debugger:
            for file, breaks in prefs.debugger.breaks.items():
                for b in breaks:
                    self.set_break(file, b)
            self.bounds, self.horpanes, self.verpanes = prefs.debugger.windowsettings
            self.tracemagic = prefs.debugger.tracemagic
        else:
            self.breaks = {}
            self.horpanes = (0.4, 0.6)
            self.verpanes = (0.3, 0.35, 0.35)
            self.bounds = (600, 400)
            self.tracemagic = 0
        self.laststacksel = None

    def canonic(self, filename):
        # override: the provided canonic() method breaks our
        # file-less Untitled windows
        return filename

    def reset(self):
        self.currentframe = None
        self.file = None
        self.laststack = None
        self.reason = 'Not running'
        self.continuewithoutdebugger = 0
        bdb.Bdb.reset(self)
        self.forget()

    def start(self, bottomframe = None, running = 0):
        W.getapplication().DebuggerQuit = bdb.BdbQuit
        from Carbon import Menu
        Menu.HiliteMenu(0)
        if self.closed:
            self.setupwidgets(self.title)
            self.closed = 0
        if not self.w.parent.debugger_quitting:
            self.w.select()
            raise W.AlertError, 'There is another debugger session busy.'
        self.reset()
        self.botframe = bottomframe
        if running:
            self.set_continue()
            self.reason = 'Running\xc9'
            self.setstate('running')
        else:
            self.set_step()
            self.reason = 'stopped'
            self.setstate('stopped')
        sys.settrace(self.trace_dispatch)

    def stop(self):
        self.set_quit()
        if self.w.parent:
            self.exit_mainloop()
            self.resetwidgets()

    def set_continue_without_debugger(self):
        sys.settrace(None)
        self.set_quit()
        self.clear_tracefuncs()
        self.continuewithoutdebugger = 1
        if hasattr(self, "w") and self.w.parent:
            self.exit_mainloop()
            self.resetwidgets()

    def clear_tracefuncs(self):
        try:
            raise 'spam'
        except:
            pass
        frame = sys.exc_traceback.tb_frame
        while frame is not None:
            del frame.f_trace
            frame = frame.f_back

    def postmortem(self, exc_type, exc_value, traceback):
        if self.closed:
            self.setupwidgets(self.title)
            self.closed = 0
        if not self.w.parent.debugger_quitting:
            raise W.AlertError, 'There is another debugger session busy.'
        self.reset()
        if traceback:
            self.botframe = traceback.tb_frame
            while traceback.tb_next <> None:
                traceback = traceback.tb_next
            frame = traceback.tb_frame
        else:
            self.botframe = None
            frame = None
        self.w.panes.bottom.buttons.killbutton.enable(1)
        self.reason = '(dead) ' + self.formatexception(exc_type, exc_value)
        self.w.select()
        self.setup(frame, traceback)
        self.setstate('dead')
        self.showstack(self.curindex)
        self.showframe(self.curindex)

    def setupwidgets(self, title):
        self.w = w = W.Window(self.bounds, title, minsize = (500, 300))

        w.panes = W.HorizontalPanes((8, 4, -8, -8), self.horpanes)

        w.panes.browserpanes = browserpanes = W.VerticalPanes(None, self.verpanes)

        browserpanes.stacklist = W.Group(None)
        browserpanes.stacklist.title = W.TextBox((4, 0, 0, 12), 'Stack')
        browserpanes.stacklist.stack = W.List((0, 16, 0, 0), callback = self.do_stack, flags = Lists.lOnlyOne)

        browserpanes.locals = W.Group(None)
        browserpanes.locals.title = W.TextBox((4, 0, 0, 12), 'Local variables')
        browserpanes.locals.browser = PyBrowser.BrowserWidget((0, 16, 0, 0))

        browserpanes.globals = W.Group(None)
        browserpanes.globals.title = W.TextBox((4, 0, 0, 12), 'Global variables')
        browserpanes.globals.browser = PyBrowser.BrowserWidget((0, 16, 0, 0))

        w.panes.bottom = bottom = W.Group(None)
        bottom.src = src = W.Group((0, 64, 0, 0))
        source = SourceViewer((1, 1, -15, -15), readonly = 1, debugger = self)
        src.optionsmenu = W.PopupMenu((-16, 0, 16, 16), [])
        src.optionsmenu.bind('<click>', self.makeoptionsmenu)

        src._barx = W.Scrollbar((0, -16, -15, 16), source.hscroll, max = 32767)
        src._bary = W.Scrollbar((-16, 15, 16, -15), source.vscroll, max = 32767)
        src.source = source
        src.frame = W.Frame((0, 0, -15, -15))

        bottom.tracingmonitor = TracingMonitor((0, 23, 6, 6))
        bottom.state = W.TextBox((12, 24, 0, 16), self.reason)

        bottom.srctitle = W.TextBox((12, 44, 0, 16))
        bottom.buttons = buttons = W.Group((12, 0, 0, 20))

        buttons.runbutton = W.Button((0, 0, 50, 16), "Run", self.do_run)
        buttons.stopbutton = W.Button((58, 0, 50, 16), "Stop", self.do_stop)
        buttons.killbutton = W.Button((116, 0, 50, 16), "Kill", self.do_kill)
        buttons.line = W.VerticalLine((173, 0, 0, 0))
        buttons.stepbutton = W.Button((181, 0, 60, 16), "Step", self.do_step)
        buttons.stepinbutton = W.Button((249, 0, 60, 16), "Step in", self.do_stepin)
        buttons.stepoutbutton = W.Button((317, 0, 60, 16), "Step out", self.do_stepout)

        w.bind('cmdr', buttons.runbutton.push)
        w.bind('cmd.', buttons.stopbutton.push)
        w.bind('cmdk', buttons.killbutton.push)
        w.bind('cmds', buttons.stepbutton.push)
        w.bind('cmdt', buttons.stepinbutton.push)
        w.bind('cmdu', buttons.stepoutbutton.push)

        w.bind('<close>', self.close)

        w.open()
        w.xxx___select(w.panes.bottom.src.source)

    def makeoptionsmenu(self):
        options = [('Clear breakpoints', self.w.panes.bottom.src.source.clearbreakpoints),
                        ('Clear all breakpoints', self.clear_all_breaks),
                        ('Edit breakpoints\xc9', self.edit_breaks), '-',
                        (self.tracemagic and
                                'Disable __magic__ tracing' or 'Enable __magic__ tracing', self.togglemagic)]
        self.w.panes.bottom.src.optionsmenu.set(options)

    def edit_breaks(self):
        if self.breaksviewer:
            self.breaksviewer.select()
        else:
            self.breaksviewer = BreakpointsViewer(self)

    def togglemagic(self):
        self.tracemagic = not self.tracemagic

    def setstate(self, state):
        self.w.panes.bottom.tracingmonitor.reset()
        self.w.panes.bottom.state.set(self.reason)
        buttons = self.w.panes.bottom.buttons
        if state == 'stopped':
            buttons.runbutton.enable(1)
            buttons.stopbutton.enable(0)
            buttons.killbutton.enable(1)
            buttons.stepbutton.enable(1)
            buttons.stepinbutton.enable(1)
            buttons.stepoutbutton.enable(1)
        elif state == 'running':
            buttons.runbutton.enable(0)
            buttons.stopbutton.enable(1)
            buttons.killbutton.enable(1)
            buttons.stepbutton.enable(0)
            buttons.stepinbutton.enable(0)
            buttons.stepoutbutton.enable(0)
        elif state == 'idle':
            buttons.runbutton.enable(0)
            buttons.stopbutton.enable(0)
            buttons.killbutton.enable(0)
            buttons.stepbutton.enable(0)
            buttons.stepinbutton.enable(0)
            buttons.stepoutbutton.enable(0)
        elif state == 'dead':
            buttons.runbutton.enable(0)
            buttons.stopbutton.enable(0)
            buttons.killbutton.enable(1)
            buttons.stepbutton.enable(0)
            buttons.stepinbutton.enable(0)
            buttons.stepoutbutton.enable(0)
        else:
            print 'unknown state:', state

    def resetwidgets(self):
        self.reason = ''
        self.w.panes.bottom.srctitle.set('')
        self.w.panes.bottom.src.source.set('')
        self.w.panes.browserpanes.stacklist.stack.set([])
        self.w.panes.browserpanes.locals.browser.set({})
        self.w.panes.browserpanes.globals.browser.set({})
        self.setstate('idle')

    # W callbacks

    def close(self):
        self.set_quit()
        self.exit_mainloop()
        self.closed = 1

        self.unregister_editor(self.w.panes.bottom.src.source,
                        self.w.panes.bottom.src.source.file)
        self.horpanes = self.w.panes.getpanesizes()
        self.verpanes = self.w.panes.browserpanes.getpanesizes()
        self.bounds = self.w.getbounds()
        prefs = W.getapplication().getprefs()
        prefs.debugger.breaks = self.breaks
        prefs.debugger.windowsettings = self.bounds, self.horpanes, self.verpanes
        prefs.debugger.tracemagic = self.tracemagic
        prefs.save()

    # stack list callback

    def do_stack(self, isdbl):
        sel = self.w.panes.browserpanes.stacklist.stack.getselection()
        if isdbl:
            if sel:
                frame, lineno = self.stack[sel[0] + 1]
                filename = frame.f_code.co_filename
                editor = self.w._parentwindow.parent.openscript(filename, lineno)
                if self.breaks.has_key(filename):
                    editor.showbreakpoints(1)
        else:
            if sel and sel <> self.laststacksel:
                self.showframe(sel[0] + 1)
            self.laststacksel = sel

    def geteditor(self, filename):
        if filename[:1] == '<' and filename[-1:] == '>':
            editor = W.getapplication().getscript(filename[1:-1])
        else:
            editor = W.getapplication().getscript(filename)
        return editor

    # button callbacks

    def do_run(self):
        self.running()
        self.set_continue()
        self.exit_mainloop()

    def do_stop(self):
        self.set_step()

    def do_kill(self):
        self.set_quit()
        self.exit_mainloop()
        self.resetwidgets()

    def do_step(self):
        self.running()
        self.set_next(self.curframe)
        self.exit_mainloop()

    def do_stepin(self):
        self.running()
        self.set_step()
        self.exit_mainloop()

    def do_stepout(self):
        self.running()
        self.set_return(self.curframe)
        self.exit_mainloop()

    def running(self):
        W.SetCursor('watch')
        self.reason = 'Running\xc9'
        self.setstate('running')
        #self.w.panes.bottom.src.source.set('')
        #self.w.panes.browserpanes.stacklist.stack.set([])
        #self.w.panes.browserpanes.locals.browser.set({})
        #self.w.panes.browserpanes.globals.browser.set({})

    def exit_mainloop(self):
        self.w.parent.debugger_quitting = 1

    #

    def showframe(self, stackindex):
        (frame, lineno) = self.stack[stackindex]
        W.SetCursor('watch')
        filename = frame.f_code.co_filename
        if filename <> self.file:
            editor = self.geteditor(filename)
            if editor:
                self.w.panes.bottom.src.source.set(editor.get(), filename)
            else:
                try:
                    f = open(filename, 'rU')
                    data = f.read()
                    f.close()
                except IOError:
                    if filename[-3:] == '.py':
                        import imp
                        modname = os.path.basename(filename)[:-3]
                        try:
                            f, filename, (suff, mode, dummy) = imp.find_module(modname)
                        except ImportError:
                            self.w.panes.bottom.src.source.set("can't find file")
                        else:
                            if f:
                                f.close()
                            if f and suff == '.py':
                                f = open(filename, 'rU')
                                data = f.read()
                                f.close()
                                self.w.panes.bottom.src.source.set(data, filename)
                            else:
                                self.w.panes.bottom.src.source.set("can't find file")
                    else:
                        self.w.panes.bottom.src.source.set("can't find file")
                else:
                    data = data.replace('\n', '\r')
                    self.w.panes.bottom.src.source.set(data, filename)
            self.file = filename
        self.w.panes.bottom.srctitle.set('Source: ' + filename + ((lineno > 0) and (' (line %d)' % lineno) or ' '))
        self.goto_line(lineno)
        self.lineno = lineno
        self.showvars((frame, lineno))

    def showvars(self, (frame, lineno)):
        if frame.f_locals is not frame.f_globals:
            locals = frame.f_locals
        else:
            locals = {'Same as Globals':''}
        filteredlocals = {}
        for key, value in locals.items():
            # empty key is magic for Python 1.4; '.' is magic for 1.5...
            if not key or key[0] <> '.':
                filteredlocals[key] = value
        self.w.panes.browserpanes.locals.browser.set(filteredlocals)
        self.w.panes.browserpanes.globals.browser.set(frame.f_globals)

    def showstack(self, stackindex):
        stack = []
        for frame, lineno in self.stack[1:]:
            filename = frame.f_code.co_filename
            try:
                filename = _filenames[filename]
            except KeyError:
                if filename[:1] + filename[-1:] <> '<>':
                    filename = os.path.basename(filename)
                _filenames[frame.f_code.co_filename] = filename
            funcname = frame.f_code.co_name
            if funcname == '?':
                funcname = '<toplevel>'
            stack.append(filename + ': ' + funcname)
        if stack <> self.laststack:
            self.w.panes.browserpanes.stacklist.stack.set(stack)
            self.laststack = stack
        sel = [stackindex - 1]
        self.w.panes.browserpanes.stacklist.stack.setselection(sel)
        self.laststacksel = sel

    def goto_line(self, lineno):
        if lineno > 0:
            self.w.panes.bottom.src.source.selectline(lineno - 1)
        else:
            self.w.panes.bottom.src.source.setselection(0, 0)

    # bdb entry points

#       def user_call(self, frame, argument_list):
#               self.reason = 'Calling'
#               self.interaction(frame, None)

    def user_line(self, frame):
        # This function is called when we stop or break at this line
        self.reason = 'Stopped'
        self.interaction(frame, None)

    def user_return(self, frame, return_value):
        # This function is called when a return trap is set here
        fname = frame.f_code.co_name
        if fname <> '?':
            self.reason = 'Returning from %s()' % frame.f_code.co_name
            frame.f_locals['__return__'] = return_value
        elif frame.f_back is self.botframe:
            self.reason = 'Done'
        else:
            self.reason = 'Returning'
        self.interaction(frame, None, 1)

    def user_exception(self, frame, (exc_type, exc_value, exc_traceback)):
        # This function is called when we stop or break at this line
        self.reason = self.formatexception(exc_type, exc_value)
        self.interaction(frame, exc_traceback)

    def formatexception(self, exc_type, exc_value):
        if exc_type == SyntaxError:
            try:
                value, (filename, lineno, charno, line) = exc_value
            except:
                pass
            else:
                return str(exc_type) + ': ' + str(value)
        if type(exc_type) == types.ClassType:
            nice = exc_type.__name__
        else:
            nice = str(exc_type)
        value = str(exc_value)
        if exc_value and value:
            nice = nice + ": " + value
        return nice

    def forget(self):
        self.stack = []
        self.curindex = 0
        self.curframe = None

    def setup(self, f, t, isreturning = 0):
        self.forget()
        self.stack, self.curindex = self.get_stack(f, t)
        self.curframe = self.stack[self.curindex - isreturning][0]

    def interaction(self, frame, traceback, isreturning = 0):
        saveport = Qd.GetPort()
        self.w.select()
        try:
            self.setup(frame, traceback, isreturning)
            self.setstate('stopped')
            stackindex = self.curindex
            if isreturning:
                if frame.f_back is not self.botframe:
                    stackindex = stackindex - 1
            self.showstack(stackindex)
            self.showframe(stackindex)
            self.w.parent.debugger_mainloop()
            self.forget()
        finally:
            Qd.SetPort(saveport)

    # bdb customization

    def trace_dispatch(self, frame, event, arg, TickCount = Evt.TickCount):
        if TickCount() - self.tracingmonitortime > 15:
            self.tracingmonitortime = TickCount()
            self.w.panes.bottom.tracingmonitor.toggle()
        try:
            try:
                if hasattr(MacOS, 'EnableAppswitch'):
                    MacOS.EnableAppswitch(0)
                if self.quitting:
                    # returning None is not enough, a former BdbQuit exception
                    # might have been eaten by the print statement
                    raise bdb.BdbQuit
                if event == 'line':
                    return self.dispatch_line(frame)
                if event == 'call':
                    return self.dispatch_call(frame, arg)
                if event == 'return':
                    return self.dispatch_return(frame, arg)
                if event == 'exception':
                    return self.dispatch_exception(frame, arg)
                print 'bdb.Bdb.dispatch: unknown debugging event:', repr(event)
                return self.trace_dispatch
            finally:
                if hasattr(MacOS, 'EnableAppswitch'):
                    MacOS.EnableAppswitch(-1)
        except KeyboardInterrupt:
            self.set_step()
            return self.trace_dispatch
        except bdb.BdbQuit:
            if self.continuewithoutdebugger:
                self.clear_tracefuncs()
                return
            else:
                raise bdb.BdbQuit
        except:
            print 'XXX Exception during debugger interaction.', \
                            self.formatexception(sys.exc_type, sys.exc_value)
            import traceback
            traceback.print_exc()
            return self.trace_dispatch

    def dispatch_call(self, frame, arg):
        if not self.tracemagic and \
                        frame.f_code.co_name[:2] == '__' == frame.f_code.co_name[-2:] and \
                        frame.f_code.co_name <> '__init__':
            return
        if self.botframe is None:
            # First call of dispatch since reset()
            self.botframe = frame.f_back    # xxx !!! added f_back
            return self.trace_dispatch
        if not (self.stop_here(frame) or self.break_anywhere(frame)):
            # No need to trace this function
            return # None
        self.user_call(frame, arg)
        if self.quitting:
            raise bdb.BdbQuit
        return self.trace_dispatch

    def set_continue(self):
        # Don't stop except at breakpoints or when finished
        self.stopframe = self.botframe
        self.returnframe = None
        self.quitting = 0
        # unlike in bdb/pdb, there's a chance that breakpoints change
        # *while* a program (this program ;-) is running. It's actually quite likely.
        # So we don't delete frame.f_trace until the bottom frame if there are no breakpoints.

    def set_break(self, filename, lineno):
        if not self.breaks.has_key(filename):
            self.breaks[filename] = []
        list = self.breaks[filename]
        if lineno in list:
            return 'There is already a breakpoint there!'
        list.append(lineno)
        list.sort()     # I want to keep them neatly sorted; easier for drawing
        if hasattr(bdb, "Breakpoint"):
            # 1.5.2b1 specific
            bp = bdb.Breakpoint(filename, lineno, 0, None)
        self.update_breaks(filename)

    def clear_break(self, filename, lineno):
        bdb.Bdb.clear_break(self, filename, lineno)
        self.update_breaks(filename)

    def clear_all_file_breaks(self, filename):
        bdb.Bdb.clear_all_file_breaks(self, filename)
        self.update_breaks(filename)

    def clear_all_breaks(self):
        bdb.Bdb.clear_all_breaks(self)
        for editors in self.editors.values():
            for editor in editors:
                editor.drawbreakpoints()

    # special

    def toggle_break(self, filename, lineno):
        if self.get_break(filename, lineno):
            self.clear_break(filename, lineno)
        else:
            self.set_break(filename, lineno)

    def clear_breaks_above(self, filename, above):
        if not self.breaks.has_key(filename):
            return 'There are no breakpoints in that file!'
        for lineno in self.breaks[filename][:]:
            if lineno > above:
                self.breaks[filename].remove(lineno)
        if not self.breaks[filename]:
            del self.breaks[filename]

    # editor stuff

    def update_breaks(self, filename):
        if self.breaksviewer:
            self.breaksviewer.update()
        if self.editors.has_key(filename):
            for editor in self.editors[filename]:
                if editor._debugger:    # XXX
                    editor.drawbreakpoints()
                else:
                    print 'xxx dead editor!'

    def update_allbreaks(self):
        if self.breaksviewer:
            self.breaksviewer.update()
        for filename in self.breaks.keys():
            if self.editors.has_key(filename):
                for editor in self.editors[filename]:
                    if editor._debugger:    # XXX
                        editor.drawbreakpoints()
                    else:
                        print 'xxx dead editor!'

    def register_editor(self, editor, filename):
        if not filename:
            return
        if not self.editors.has_key(filename):
            self.editors[filename] = [editor]
        elif editor not in self.editors[filename]:
            self.editors[filename].append(editor)

    def unregister_editor(self, editor, filename):
        if not filename:
            return
        try:
            self.editors[filename].remove(editor)
            if not self.editors[filename]:
                del self.editors[filename]
                # if this was an untitled window, clear the breaks.
                if filename[:1] == '<' and filename[-1:] == '>' and \
                                self.breaks.has_key(filename):
                    self.clear_all_file_breaks(filename)
        except (KeyError, ValueError):
            pass


class SourceViewer(W.PyEditor):

    def __init__(self, *args, **kwargs):
        apply(W.PyEditor.__init__, (self,) + args, kwargs)
        self.bind('<click>', self.clickintercept)

    def clickintercept(self, point, modifiers):
        if self._parentwindow._currentwidget <> self and not self.pt_in_breaks(point):
            self._parentwindow.xxx___select(self)
            return 1

    def _getviewrect(self):
        l, t, r, b = self._bounds
        if self._debugger:
            return (l + 12, t + 2, r - 1, b - 2)
        else:
            return (l + 5, t + 2, r - 1, b - 2)

    def select(self, onoff, isclick = 0):
        if W.SelectableWidget.select(self, onoff):
            return
        self.SetPort()
        #if onoff:
        #       self.ted.WEActivate()
        #else:
        #       self.ted.WEDeactivate()
        self.drawselframe(onoff)

    def drawselframe(self, onoff):
        pass


class BreakpointsViewer:

    def __init__(self, debugger):
        self.debugger = debugger
        self.w = W.Window((300, 250), 'Breakpoints', minsize = (200, 200))
        self.w.panes = W.HorizontalPanes((8, 8, -8, -32), (0.3, 0.7))
        self.w.panes.files = W.List(None, callback = self.filehit)              #, flags = Lists.lOnlyOne)
        self.w.panes.gr = W.Group(None)
        self.w.panes.gr.breaks = W.List((0, 0, -130, 0), callback = self.linehit)       #, flags = Lists.lOnlyOne)
        self.w.panes.gr.openbutton = W.Button((-80, 4, 0, 16), 'View\xc9', self.openbuttonhit)
        self.w.panes.gr.deletebutton = W.Button((-80, 28, 0, 16), 'Delete', self.deletebuttonhit)

        self.w.bind('<close>', self.close)
        self.w.bind('backspace', self.w.panes.gr.deletebutton.push)

        self.setup()
        self.w.open()
        self.w.panes.gr.openbutton.enable(0)
        self.w.panes.gr.deletebutton.enable(0)
        self.curfile = None

    def deletebuttonhit(self):
        if self.w._currentwidget == self.w.panes.files:
            self.del_filename()
        else:
            self.del_number()
        self.checkbuttons()

    def del_number(self):
        if self.curfile is None:
            return
        sel = self.w.panes.gr.breaks.getselectedobjects()
        for lineno in sel:
            self.debugger.clear_break(self.curfile, lineno)

    def del_filename(self):
        sel = self.w.panes.files.getselectedobjects()
        for filename in sel:
            self.debugger.clear_all_file_breaks(filename)
        self.debugger.update_allbreaks()

    def setup(self):
        files = self.debugger.breaks.keys()
        files.sort()
        self.w.panes.files.set(files)

    def close(self):
        self.debugger.breaksviewer = None
        self.debugger = None

    def update(self):
        sel = self.w.panes.files.getselectedobjects()
        self.setup()
        self.w.panes.files.setselectedobjects(sel)
        sel = self.w.panes.files.getselection()
        if len(sel) == 0 and self.curfile:
            self.w.panes.files.setselectedobjects([self.curfile])
        self.filehit(0)

    def select(self):
        self.w.select()

    def selectfile(self, file):
        self.w.panes.files.setselectedobjects([file])
        self.filehit(0)

    def openbuttonhit(self):
        self.filehit(1)

    def filehit(self, isdbl):
        sel = self.w.panes.files.getselectedobjects()
        if isdbl:
            for filename in sel:
                lineno = None
                if filename == self.curfile:
                    linesel = self.w.panes.gr.breaks.getselectedobjects()
                    if linesel:
                        lineno = linesel[-1]
                    elif self.w.panes.gr.breaks:
                        lineno = self.w.panes.gr.breaks[0]
                editor = self.w._parentwindow.parent.openscript(filename, lineno)
                editor.showbreakpoints(1)
            return
        if len(sel) == 1:
            file = sel[0]
            filebreaks = self.debugger.breaks[file][:]
            if self.curfile == file:
                linesel = self.w.panes.gr.breaks.getselectedobjects()
            self.w.panes.gr.breaks.set(filebreaks)
            if self.curfile == file:
                self.w.panes.gr.breaks.setselectedobjects(linesel)
            self.curfile = file
        else:
            if len(sel) <> 0:
                self.curfile = None
            self.w.panes.gr.breaks.set([])
        self.checkbuttons()

    def linehit(self, isdbl):
        if isdbl:
            files = self.w.panes.files.getselectedobjects()
            if len(files) <> 1:
                return
            filename = files[0]
            linenos = self.w.panes.gr.breaks.getselectedobjects()
            if not linenos:
                return
            lineno = linenos[-1]
            editor = self.w._parentwindow.parent.openscript(filename, lineno)
            editor.showbreakpoints(1)
        self.checkbuttons()

    def checkbuttons(self):
        if self.w.panes.files.getselection():
            self.w.panes.gr.openbutton.enable(1)
            self.w._parentwindow.setdefaultbutton(self.w.panes.gr.openbutton)
            if self.w._currentwidget == self.w.panes.files:
                if self.w.panes.files.getselection():
                    self.w.panes.gr.deletebutton.enable(1)
                else:
                    self.w.panes.gr.deletebutton.enable(0)
            else:
                if self.w.panes.gr.breaks.getselection():
                    self.w.panes.gr.deletebutton.enable(1)
                else:
                    self.w.panes.gr.deletebutton.enable(0)
        else:
            self.w.panes.gr.openbutton.enable(0)
            self.w.panes.gr.deletebutton.enable(0)


class TracingMonitor(W.Widget):

    def __init__(self, *args, **kwargs):
        apply(W.Widget.__init__, (self,) + args, kwargs)
        self.state = 0

    def toggle(self):
        if hasattr(self, "_parentwindow") and self._parentwindow is not None:
            self.state = self.state % 2 + 1
            port = Qd.GetPort()
            self.SetPort()
            self.draw()
            Qd.SetPort(port)

    def reset(self):
        if self._parentwindow:
            self.state = 0
            port = Qd.GetPort()
            self.SetPort()
            self.draw()
            Qd.SetPort(port)

    def draw(self, visRgn = None):
        if self.state == 2:
            Qd.PaintOval(self._bounds)
        else:
            Qd.EraseOval(self._bounds)


# convenience funcs

def postmortem(exc_type, exc_value, tb):
    d = getdebugger()
    d.postmortem(exc_type, exc_value, tb)

def start(bottomframe = None):
    d = getdebugger()
    d.start(bottomframe)

def startfromhere():
    d = getdebugger()
    try:
        raise 'spam'
    except:
        frame = sys.exc_traceback.tb_frame.f_back
    d.start(frame)

def startfrombottom():
    d = getdebugger()
    d.start(_getbottomframe(), 1)

def stop():
    d = getdebugger()
    d.stop()

def cont():
    sys.settrace(None)
    d = getdebugger()
    d.set_continue_without_debugger()

def _getbottomframe():
    try:
        raise 'spam'
    except:
        pass
    frame = sys.exc_traceback.tb_frame
    while 1:
        if frame.f_code.co_name == 'mainloop' or frame.f_back is None:
            break
        frame = frame.f_back
    return frame

_debugger = None

def getdebugger():
    if not __debug__:
        raise W.AlertError, "Can't debug in \"Optimize bytecode\" mode.\r(see \"Default startup options\" in EditPythonPreferences)"
    global _debugger
    if _debugger is None:
        _debugger = Debugger()
    return _debugger



# Copyright (C) 2001-2004 Python Software Foundation
# Author: Barry Warsaw, Thomas Wouters, Anthony Baxter
# Contact: email-sig@python.org

"""A parser of RFC 2822 and MIME email messages."""

import re
from cStringIO import StringIO
from email.FeedParser import FeedParser
from email.Message import Message

NLCRE = re.compile('\r\n|\r|\n')



class Parser:
    def __init__(self, _class=Message, strict=False):
        """Parser of RFC 2822 and MIME email messages.

        Creates an in-memory object tree representing the email message, which
        can then be manipulated and turned over to a Generator to return the
        textual representation of the message.

        The string must be formatted as a block of RFC 2822 headers and header
        continuation lines, optionally preceeded by a `Unix-from' header.  The
        header block is terminated either by the end of the string or by a
        blank line.

        _class is the class to instantiate for new message objects when they
        must be created.  This class must have a constructor that can take
        zero arguments.  Default is Message.Message.

        Optional strict tells the parser to be strictly RFC compliant or to be
        more forgiving in parsing of ill-formatted MIME documents.  When
        non-strict mode is used, the parser will try to make up for missing or
        erroneous boundaries and other peculiarities seen in the wild.
        Default is non-strict parsing.
        """
        self._class = _class

    def parse(self, fp, headersonly=False):
        """Create a message structure from the data in a file.

        Reads all the data from the file and returns the root of the message
        structure.  Optional headersonly is a flag specifying whether to stop
        parsing after reading the headers or not.  The default is False,
        meaning it parses the entire contents of the file.
        """
        feedparser = FeedParser(self._class)
        if headersonly:
            feedparser._set_headersonly()
        while True:
            data = fp.read(8192)
            if not data:
                break
            feedparser.feed(data)
        return feedparser.close()

    def parsestr(self, text, headersonly=False):
        """Create a message structure from a string.

        Returns the root of the message structure.  Optional headersonly is a
        flag specifying whether to stop parsing after reading the headers or
        not.  The default is False, meaning it parses the entire contents of
        the file.
        """
        return self.parse(StringIO(text), headersonly=headersonly)



class HeaderParser(Parser):
    def parse(self, fp, headersonly=True):
        return Parser.parse(self, fp, True)

    def parsestr(self, text, headersonly=True):
        return Parser.parsestr(self, text, True)



# Text formatting abstractions
# Note -- this module is obsolete, it's too slow anyway


# Oft-used type object
Int = type(0)


# Represent a paragraph.  This is a list of words with associated
# font and size information, plus indents and justification for the
# entire paragraph.
# Once the words have been added to a paragraph, it can be laid out
# for different line widths.  Once laid out, it can be rendered at
# different screen locations.  Once rendered, it can be queried
# for mouse hits, and parts of the text can be highlighted
class Para:
    #
    def __init__(self):
        self.words = [] # The words
        self.just = 'l' # Justification: 'l', 'r', 'lr' or 'c'
        self.indent_left = self.indent_right = self.indent_hang = 0
        # Final lay-out parameters, may change
        self.left = self.top = self.right = self.bottom = \
                self.width = self.height = self.lines = None
    #
    # Add a word, computing size information for it.
    # Words may also be added manually by appending to self.words
    # Each word should be a 7-tuple:
    # (font, text, width, space, stretch, ascent, descent)
    def addword(self, d, font, text, space, stretch):
        if font is not None:
            d.setfont(font)
        width = d.textwidth(text)
        ascent = d.baseline()
        descent = d.lineheight() - ascent
        spw = d.textwidth(' ')
        space = space * spw
        stretch = stretch * spw
        tuple = (font, text, width, space, stretch, ascent, descent)
        self.words.append(tuple)
    #
    # Hooks to begin and end anchors -- insert numbers in the word list!
    def bgn_anchor(self, id):
        self.words.append(id)
    #
    def end_anchor(self, id):
        self.words.append(0)
    #
    # Return the total length (width) of the text added so far, in pixels
    def getlength(self):
        total = 0
        for word in self.words:
            if type(word) is not Int:
                total = total + word[2] + word[3]
        return total
    #
    # Tab to a given position (relative to the current left indent):
    # remove all stretch, add fixed space up to the new indent.
    # If the current position is already at the tab stop,
    # don't add any new space (but still remove the stretch)
    def tabto(self, tab):
        total = 0
        as, de = 1, 0
        for i in range(len(self.words)):
            word = self.words[i]
            if type(word) is Int: continue
            (fo, te, wi, sp, st, as, de) = word
            self.words[i] = (fo, te, wi, sp, 0, as, de)
            total = total + wi + sp
        if total < tab:
            self.words.append((None, '', 0, tab-total, 0, as, de))
    #
    # Make a hanging tag: tab to hang, increment indent_left by hang,
    # and reset indent_hang to -hang
    def makehangingtag(self, hang):
        self.tabto(hang)
        self.indent_left = self.indent_left + hang
        self.indent_hang = -hang
    #
    # Decide where the line breaks will be given some screen width
    def layout(self, linewidth):
        self.width = linewidth
        height = 0
        self.lines = lines = []
        avail1 = self.width - self.indent_left - self.indent_right
        avail = avail1 - self.indent_hang
        words = self.words
        i = 0
        n = len(words)
        lastfont = None
        while i < n:
            firstfont = lastfont
            charcount = 0
            width = 0
            stretch = 0
            ascent = 0
            descent = 0
            lsp = 0
            j = i
            while i < n:
                word = words[i]
                if type(word) is Int:
                    if word > 0 and width >= avail:
                        break
                    i = i+1
                    continue
                fo, te, wi, sp, st, as, de = word
                if width + wi > avail and width > 0 and wi > 0:
                    break
                if fo is not None:
                    lastfont = fo
                    if width == 0:
                        firstfont = fo
                charcount = charcount + len(te) + (sp > 0)
                width = width + wi + sp
                lsp = sp
                stretch = stretch + st
                lst = st
                ascent = max(ascent, as)
                descent = max(descent, de)
                i = i+1
            while i > j and type(words[i-1]) is Int and \
                    words[i-1] > 0: i = i-1
            width = width - lsp
            if i < n:
                stretch = stretch - lst
            else:
                stretch = 0
            tuple = i-j, firstfont, charcount, width, stretch, \
                    ascent, descent
            lines.append(tuple)
            height = height + ascent + descent
            avail = avail1
        self.height = height
    #
    # Call a function for all words in a line
    def visit(self, wordfunc, anchorfunc):
        avail1 = self.width - self.indent_left - self.indent_right
        avail = avail1 - self.indent_hang
        v = self.top
        i = 0
        for tuple in self.lines:
            wordcount, firstfont, charcount, width, stretch, \
                    ascent, descent = tuple
            h = self.left + self.indent_left
            if i == 0: h = h + self.indent_hang
            extra = 0
            if self.just == 'r': h = h + avail - width
            elif self.just == 'c': h = h + (avail - width) / 2
            elif self.just == 'lr' and stretch > 0:
                extra = avail - width
            v2 = v + ascent + descent
            for j in range(i, i+wordcount):
                word = self.words[j]
                if type(word) is Int:
                    ok = anchorfunc(self, tuple, word, \
                                    h, v)
                    if ok is not None: return ok
                    continue
                fo, te, wi, sp, st, as, de = word
                if extra > 0 and stretch > 0:
                    ex = extra * st / stretch
                    extra = extra - ex
                    stretch = stretch - st
                else:
                    ex = 0
                h2 = h + wi + sp + ex
                ok = wordfunc(self, tuple, word, h, v, \
                        h2, v2, (j==i), (j==i+wordcount-1))
                if ok is not None: return ok
                h = h2
            v = v2
            i = i + wordcount
            avail = avail1
    #
    # Render a paragraph in "drawing object" d, using the rectangle
    # given by (left, top, right) with an unspecified bottom.
    # Return the computed bottom of the text.
    def render(self, d, left, top, right):
        if self.width != right-left:
            self.layout(right-left)
        self.left = left
        self.top = top
        self.right = right
        self.bottom = self.top + self.height
        self.anchorid = 0
        try:
            self.d = d
            self.visit(self.__class__._renderword, \
                       self.__class__._renderanchor)
        finally:
            self.d = None
        return self.bottom
    #
    def _renderword(self, tuple, word, h, v, h2, v2, isfirst, islast):
        if word[0] is not None: self.d.setfont(word[0])
        baseline = v + tuple[5]
        self.d.text((h, baseline - word[5]), word[1])
        if self.anchorid > 0:
            self.d.line((h, baseline+2), (h2, baseline+2))
    #
    def _renderanchor(self, tuple, word, h, v):
        self.anchorid = word
    #
    # Return which anchor(s) was hit by the mouse
    def hitcheck(self, mouseh, mousev):
        self.mouseh = mouseh
        self.mousev = mousev
        self.anchorid = 0
        self.hits = []
        self.visit(self.__class__._hitcheckword, \
                   self.__class__._hitcheckanchor)
        return self.hits
    #
    def _hitcheckword(self, tuple, word, h, v, h2, v2, isfirst, islast):
        if self.anchorid > 0 and h <= self.mouseh <= h2 and \
                v <= self.mousev <= v2:
            self.hits.append(self.anchorid)
    #
    def _hitcheckanchor(self, tuple, word, h, v):
        self.anchorid = word
    #
    # Return whether the given anchor id is present
    def hasanchor(self, id):
        return id in self.words or -id in self.words
    #
    # Extract the raw text from the word list, substituting one space
    # for non-empty inter-word space, and terminating with '\n'
    def extract(self):
        text = ''
        for w in self.words:
            if type(w) is not Int:
                word = w[1]
                if w[3]: word = word + ' '
                text = text + word
        return text + '\n'
    #
    # Return which character position was hit by the mouse, as
    # an offset in the entire text as returned by extract().
    # Return None if the mouse was not in this paragraph
    def whereis(self, d, mouseh, mousev):
        if mousev < self.top or mousev > self.bottom:
            return None
        self.mouseh = mouseh
        self.mousev = mousev
        self.lastfont = None
        self.charcount = 0
        try:
            self.d = d
            return self.visit(self.__class__._whereisword, \
                              self.__class__._whereisanchor)
        finally:
            self.d = None
    #
    def _whereisword(self, tuple, word, h1, v1, h2, v2, isfirst, islast):
        fo, te, wi, sp, st, as, de = word
        if fo is not None: self.lastfont = fo
        h = h1
        if isfirst: h1 = 0
        if islast: h2 = 999999
        if not (v1 <= self.mousev <= v2 and h1 <= self.mouseh <= h2):
            self.charcount = self.charcount + len(te) + (sp > 0)
            return
        if self.lastfont is not None:
            self.d.setfont(self.lastfont)
        cc = 0
        for c in te:
            cw = self.d.textwidth(c)
            if self.mouseh <= h + cw/2:
                return self.charcount + cc
            cc = cc+1
            h = h+cw
        self.charcount = self.charcount + cc
        if self.mouseh <= (h+h2) / 2:
            return self.charcount
        else:
            return self.charcount + 1
    #
    def _whereisanchor(self, tuple, word, h, v):
        pass
    #
    # Return screen position corresponding to position in paragraph.
    # Return tuple (h, vtop, vbaseline, vbottom).
    # This is more or less the inverse of whereis()
    def screenpos(self, d, pos):
        if pos < 0:
            ascent, descent = self.lines[0][5:7]
            return self.left, self.top, self.top + ascent, \
                    self.top + ascent + descent
        self.pos = pos
        self.lastfont = None
        try:
            self.d = d
            ok = self.visit(self.__class__._screenposword, \
                            self.__class__._screenposanchor)
        finally:
            self.d = None
        if ok is None:
            ascent, descent = self.lines[-1][5:7]
            ok = self.right, self.bottom - ascent - descent, \
                    self.bottom - descent, self.bottom
        return ok
    #
    def _screenposword(self, tuple, word, h1, v1, h2, v2, isfirst, islast):
        fo, te, wi, sp, st, as, de = word
        if fo is not None: self.lastfont = fo
        cc = len(te) + (sp > 0)
        if self.pos > cc:
            self.pos = self.pos - cc
            return
        if self.pos < cc:
            self.d.setfont(self.lastfont)
            h = h1 + self.d.textwidth(te[:self.pos])
        else:
            h = h2
        ascent, descent = tuple[5:7]
        return h, v1, v1+ascent, v2
    #
    def _screenposanchor(self, tuple, word, h, v):
        pass
    #
    # Invert the stretch of text between pos1 and pos2.
    # If pos1 is None, the beginning is implied;
    # if pos2 is None, the end is implied.
    # Undoes its own effect when called again with the same arguments
    def invert(self, d, pos1, pos2):
        if pos1 is None:
            pos1 = self.left, self.top, self.top, self.top
        else:
            pos1 = self.screenpos(d, pos1)
        if pos2 is None:
            pos2 = self.right, self.bottom,self.bottom,self.bottom
        else:
            pos2 = self.screenpos(d, pos2)
        h1, top1, baseline1, bottom1 = pos1
        h2, top2, baseline2, bottom2 = pos2
        if bottom1 <= top2:
            d.invert((h1, top1), (self.right, bottom1))
            h1 = self.left
            if bottom1 < top2:
                d.invert((h1, bottom1), (self.right, top2))
            top1, bottom1 = top2, bottom2
        d.invert((h1, top1), (h2, bottom2))



import W

# this demoscript illustrates how to tie a callback to activating or clicking away of the window.
# evb 22 4 99

class ActivateDemo:

    def __init__(self):
        self.w = W.Window((200, 200), 'Activate demo')
        self.w.bind("<activate>", self.my_activate_callback)
        self.w.open()

    def my_activate_callback(self, onoff):
        '''the callback gets 1 parameter which indicates whether the window
        has been activated (1) or clicked to the back (0)'''
        if onoff == 1:
            print "I'm in the front now!"
        else:
            print "I've been clicked away, Oh No!"

ad = ActivateDemo()



# List all resources

from Carbon import Res
from Carbon.Resources import *

def list1resources():
    ntypes = Res.Count1Types()
    for itype in range(1, 1+ntypes):
        type = Res.Get1IndType(itype)
        print "Type:", repr(type)
        nresources = Res.Count1Resources(type)
        for i in range(1, 1 + nresources):
            Res.SetResLoad(0)
            res = Res.Get1IndResource(type, i)
            Res.SetResLoad(1)
            info(res)

def listresources():
    ntypes = Res.CountTypes()
    for itype in range(1, 1+ntypes):
        type = Res.GetIndType(itype)
        print "Type:", repr(type)
        nresources = Res.CountResources(type)
        for i in range(1, 1 + nresources):
            Res.SetResLoad(0)
            res = Res.GetIndResource(type, i)
            Res.SetResLoad(1)
            info(res)

def info(res):
    print res.GetResInfo(), res.SizeResource(), decodeattrs(res.GetResAttrs())

attrnames = {
        resChanged:     'Changed',
        resPreload:     'Preload',
        resProtected:   'Protected',
        resLocked:      'Locked',
        resPurgeable:   'Purgeable',
        resSysHeap:     'SysHeap',
}

def decodeattrs(attrs):
    names = []
    for bit in range(16):
        mask = 1<<bit
        if attrs & mask:
            if attrnames.has_key(mask):
                names.append(attrnames[mask])
            else:
                names.append(hex(mask))
    return names

def test():
    print "=== Local resourcess ==="
    list1resources()
    print "=== All resources ==="
    listresources()

if __name__ == '__main__':
    test()



# Scan an Apple header file, generating a Python file of generator calls.

import sys
import os
from bgenlocations import TOOLBOXDIR, BGENDIR
sys.path.append(BGENDIR)
from scantools import Scanner_OSX

LONG = "CoreFoundation"
SHORT = "cf"
OBJECTS = ("CFTypeRef",
                "CFArrayRef", "CFMutableArrayRef",
                "CFDataRef", "CFMutableDataRef",
                "CFDictionaryRef", "CFMutableDictionaryRef",
                "CFStringRef", "CFMutableStringRef",
                "CFURLRef",
##              "CFPropertyListRef",
                )
# ADD object typenames here

def main():
    input = [
            "CFBase.h",
            "CFArray.h",
##              "CFBag.h",
##              "CFBundle.h",
##              "CFCharacterSet.h",
            "CFData.h",
##              "CFDate.h",
            "CFDictionary.h",
##              "CFNumber.h",
##              "CFPlugIn.h",
            "CFPreferences.h",
            "CFPropertyList.h",
##              "CFSet.h",
            "CFString.h",
##              "CFStringEncodingExt.h",
##              "CFTimeZone.h",
            "CFURL.h",
            ]
    output = SHORT + "gen.py"
    defsoutput = TOOLBOXDIR + LONG + ".py"
    scanner = MyScanner(input, output, defsoutput)
    scanner.scan()
    scanner.gentypetest(SHORT+"typetest.py")
    scanner.close()
    print "=== Testing definitions output code ==="
    execfile(defsoutput, {}, {})
    print "=== Done scanning and generating, now importing the generated code... ==="
    exec "import " + SHORT + "support"
    print "=== Done.  It's up to you to compile it now! ==="

class MyScanner(Scanner_OSX):

    def destination(self, type, name, arglist):
        classname = "Function"
        listname = "functions"
        if arglist and name[:13] != 'CFPreferences':
            t, n, m = arglist[0]
            if t in OBJECTS and m == "InMode":
                classname = "Method"
                listname = t + "_methods"
            # Special case for the silly first AllocatorRef argument
            if t == 'CFAllocatorRef' and m == 'InMode' and len(arglist) > 1:
                t, n, m = arglist[1]
                if t in OBJECTS and m == "InMode":
                    classname = "MethodSkipArg1"
                    listname = t + "_methods"
        return classname, listname

    def writeinitialdefs(self):
        self.defsfile.write("def FOUR_CHAR_CODE(x): return x\n")

    def makeblacklistnames(self):
        return [
                # Memory allocator functions
                "CFAllocatorGetDefault",
                "CFAllocatorSetDefault",
                "CFAllocatorAllocate",
                "CFAllocatorReallocate",
                "CFAllocatorDeallocate",
                "CFGetAllocator",
                # Array functions we skip for now.
                "CFArrayGetValueAtIndex",
                # Data pointer functions. Skip for now.
                "CFDataGetBytePtr",
                "CFDataGetMutableBytePtr",
                "CFDataGetBytes",   # XXXX Should support this one
                # String functions
                "CFStringGetPascalString", # Use the C-string methods.
                "CFStringGetPascalStringPtr", # TBD automatically
                "CFStringGetCStringPtr",
                "CFStringGetCharactersPtr",
                "CFStringGetCString",
                "CFStringGetCharacters",
                "CFURLCreateStringWithFileSystemPath", # Gone in later releases
                "CFStringCreateMutableWithExternalCharactersNoCopy", # Not a clue...
                "CFStringSetExternalCharactersNoCopy",
                "CFStringGetCharacterAtIndex", # No format for single unichars yet.
                "kCFStringEncodingInvalidId", # incompatible constant declaration
                "CFPropertyListCreateFromXMLData", # Manually generated
                ]

    def makegreylist(self):
        return []

    def makeblacklisttypes(self):
        return [
                "CFComparatorFunction", # Callback function pointer
                "CFAllocatorContext", # Not interested in providing our own allocator
                "void_ptr_ptr",  # Tricky. This is the initializer for arrays...
                "void_ptr", # Ditto for various array lookup methods
                "CFArrayApplierFunction", # Callback function pointer
                "CFDictionaryApplierFunction", # Callback function pointer
                "va_list", # For printf-to-a-cfstring. Use Python.
                "const_CFStringEncoding_ptr", # To be done, I guess
                ]

    def makerepairinstructions(self):
        return [
                # Buffers in CF seem to be passed as UInt8 * normally.
                ([("UInt8_ptr", "*", "InMode"), ("CFIndex", "*", "InMode")],
                 [("UcharInBuffer", "*", "*")]),

                ([("UniChar_ptr", "*", "InMode"), ("CFIndex", "*", "InMode")],
                 [("UnicodeInBuffer", "*", "*")]),

                # Some functions return a const char *. Don't worry, we won't modify it.
                ([("const_char_ptr", "*", "ReturnMode")],
                 [("return_stringptr", "*", "*")]),

                # base URLs are optional (pass None for NULL)
                ([("CFURLRef", "baseURL", "InMode")],
                 [("OptionalCFURLRef", "*", "*")]),

                # We handle CFPropertyListRef objects as plain CFTypeRef
                ([("CFPropertyListRef", "*", "*")],
                 [("CFTypeRef", "*", "*")]),
                ]

if __name__ == "__main__":
    main()



import Wbase
import Wkeys
import string
from Carbon import Evt, Events, Fm, Lists, Qd, Scrap, Win
from Carbon.List import LNew, CreateCustomList
from Carbon.Lists import kListDefUserProcType, lInitMsg, lDrawMsg, lHiliteMsg, lCloseMsg
from Carbon.QuickDraw import hilitetransfermode
from Carbon import App
from Carbon.Appearance import kThemeStateActive, kThemeStateInactive, kThemeStatePressed


class List(Wbase.SelectableWidget):

    """Standard list widget."""

    LDEF_ID = 0

    def __init__(self, possize, items = None, callback = None, flags = 0, cols = 1, typingcasesens=0):
        if items is None:
            items = []
        self.items = items
        Wbase.SelectableWidget.__init__(self, possize)
        self._selected = 0
        self._enabled = 1
        self._list = None
        self._cols = cols
        self._callback = callback
        self._flags = flags
        self.typingcasesens = typingcasesens
        self.lasttyping = ""
        self.lasttime = Evt.TickCount()
        self.timelimit = 30
        self.setitems(items)
        self.drawingmode = 0

    def open(self):
        self.setdrawingmode(0)
        self.createlist()
        self.setdrawingmode(1)

    def createlist(self):
        self._calcbounds()
        self.SetPort()
        rect = self._bounds
        rect = rect[0]+1, rect[1]+1, rect[2]-16, rect[3]-1
        self._viewbounds = rect
        self._list = LNew(rect, (0, 0, self._cols, 0), (0, 0), self.LDEF_ID, self._parentwindow.wid,
                                0, 1, 0, 1)
        if self.drawingmode:
            self._list.LSetDrawingMode(0)
        self._list.selFlags = self._flags
        self.setitems(self.items)
        if hasattr(self, "_sel"):
            self.setselection(self._sel)
            del self._sel

    def adjust(self, oldbounds):
        self.SetPort()
        # Appearance frames are drawn outside the specified bounds,
        # so we always need to outset the invalidated area.
        self.GetWindow().InvalWindowRect(Qd.InsetRect(oldbounds, -3, -3))
        self.GetWindow().InvalWindowRect(Qd.InsetRect(self._bounds, -3, -3))

        if oldbounds[:2] == self._bounds[:2]:
            # set visRgn to empty, to prevent nasty drawing side effect of LSize()
            Qd.RectRgn(self._parentwindow.wid.GetWindowPort().visRgn, (0, 0, 0, 0))
            # list still has the same upper/left coordinates, use LSize
            l, t, r, b = self._bounds
            width = r - l - 17
            height = b - t - 2
            vl, vt, vr, vb = self._viewbounds
            self._viewbounds = vl, vt, vl + width, vt + height
            self._list.LSize(width, height)
            # now *why* doesn't the list manager recalc the cellrect???
            l, t, r, b = self._list.LRect((0,0))
            cellheight = b - t
            self._list.LCellSize((width/self._cols, cellheight))
            # reset visRgn
            self._parentwindow.wid.CalcVis()
        else:
            # oh well, since the list manager doesn't have a LMove call,
            # we have to make the list all over again...
            sel = self.getselection()
            topcell = self.gettopcell()
            self._list = None
            self.setdrawingmode(0)
            self.createlist()
            self.setselection(sel)
            self.settopcell(topcell)
            self.setdrawingmode(1)

    def close(self):
        self._list = None
        self._callback = None
        self.items = []
        Wbase.SelectableWidget.close(self)

    def set(self, items):
        self.setitems(items)

    def setitems(self, items):
        self.items = items
        the_list = self._list
        if not self._parent or not self._list:
            return
        self.setdrawingmode(0)
        topcell = self.gettopcell()
        the_list.LDelRow(0, 1)
        the_list.LAddRow(len(self.items), 0)
        self_itemrepr = self.itemrepr
        set_cell = the_list.LSetCell
        for i in range(len(items)):
            set_cell(self_itemrepr(items[i]), (0, i))
        self.settopcell(topcell)
        self.setdrawingmode(1)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        isdoubleclick = self._list.LClick(point, modifiers)
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, isdoubleclick)
        return 1

    def key(self, char, event):
        (what, message, when, where, modifiers) = event
        sel = self.getselection()
        newselection = []
        if char == Wkeys.uparrowkey:
            if len(sel) >= 1 and min(sel) > 0:
                newselection = [min(sel) - 1]
            else:
                newselection = [0]
        elif char == Wkeys.downarrowkey:
            if len(sel) >= 1 and max(sel) < (len(self.items) - 1):
                newselection = [max(sel) + 1]
            else:
                newselection = [len(self.items) - 1]
        else:
            modifiers = 0
            if (self.lasttime + self.timelimit) < Evt.TickCount():
                self.lasttyping = ""
            if self.typingcasesens:
                self.lasttyping = self.lasttyping + char
            else:
                self.lasttyping = self.lasttyping + string.lower(char)
            self.lasttime = Evt.TickCount()
            i = self.findmatch(self.lasttyping)
            newselection = [i]
        if modifiers & Events.shiftKey and not self._list.selFlags & Lists.lOnlyOne:
            newselection = newselection + sel
        self.setselection(newselection)
        self._list.LAutoScroll()
        self.click((-1, -1), 0)

    def findmatch(self, tag):
        lower = string.lower
        items = self.items
        typingcasesens = self.typingcasesens
        taglen = len(tag)
        match = '\377' * 100
        match_i = -1
        for i in range(len(items)):
            item = str(items[i])
            if not typingcasesens:
                item = lower(item)
            if tag <= item < match:
                match = item
                match_i = i
        if match_i >= 0:
            return match_i
        else:
            return len(items) - 1

    def domenu_copy(self, *args):
        sel = self.getselection()
        selitems = []
        for i in sel:
            selitems.append(str(self.items[i]))
        text = string.join(selitems, '\r')
        if text:
            if hasattr(Scrap, 'PutScrap'):
                Scrap.ZeroScrap()
                Scrap.PutScrap('TEXT', text)
            else:
                Scrap.ClearCurrentScrap()
                sc = Scrap.GetCurrentScrap()
                sc.PutScrapFlavor('TEXT', 0, text)

    def can_copy(self, *args):
        return len(self.getselection()) <> 0

    def domenu_selectall(self, *args):
        self.selectall()

    def can_selectall(self, *args):
        return not self._list.selFlags & Lists.lOnlyOne

    def selectall(self):
        if not self._list.selFlags & Lists.lOnlyOne:
            self.setselection(range(len(self.items)))
            self._list.LAutoScroll()
            self.click((-1, -1), 0)

    def getselection(self):
        if not self._parent or not self._list:
            if hasattr(self, "_sel"):
                return self._sel
            return []
        items = []
        point = (0,0)
        while 1:
            ok, point = self._list.LGetSelect(1, point)
            if not ok:
                break
            items.append(point[1])
            point = point[0], point[1]+1
        return items

    def setselection(self, selection):
        if not self._parent or not self._list:
            self._sel = selection
            return
        set_sel = self._list.LSetSelect
        for i in range(len(self.items)):
            if i in selection:
                set_sel(1, (0, i))
            else:
                set_sel(0, (0, i))
        self._list.LAutoScroll()

    def getselectedobjects(self):
        sel = self.getselection()
        objects = []
        for i in sel:
            objects.append(self.items[i])
        return objects

    def setselectedobjects(self, objects):
        sel = []
        for o in objects:
            try:
                sel.append(self.items.index(o))
            except:
                pass
        self.setselection(sel)

    def gettopcell(self):
        l, t, r, b = self._bounds
        t = t + 1
        cl, ct, cr, cb = self._list.LRect((0, 0))
        cellheight = cb - ct
        return (t - ct) / cellheight

    def settopcell(self, topcell):
        top = self.gettopcell()
        diff = topcell - top
        self._list.LScroll(0, diff)

    def draw(self, visRgn = None):
        if self._visible:
            if not visRgn:
                visRgn = self._parentwindow.wid.GetWindowPort().visRgn
            self._list.LUpdate(visRgn)
            state = [kThemeStateActive, kThemeStateInactive][not self._activated]
            App.DrawThemeListBoxFrame(Qd.InsetRect(self._bounds, 1, 1), state)
            if self._selected and self._activated:
                self.drawselframe(1)

    def select(self, onoff, isclick = 0):
        if Wbase.SelectableWidget.select(self, onoff):
            return
        self.SetPort()
        self.drawselframe(onoff)

    def activate(self, onoff):
        self._activated = onoff
        if self._visible:
            self._list.LActivate(onoff)
            #state = [kThemeStateActive, kThemeStateInactive][not onoff]
            #App.DrawThemeListBoxFrame(Qd.InsetRect(self._bounds, 1, 1), state)
            if self._selected:
                self.drawselframe(onoff)

    def get(self):
        return self.items

    def itemrepr(self, item):
        return str(item)[:255]

    def __getitem__(self, index):
        return self.items[index]

    def __setitem__(self, index, item):
        if self._parent and self._list:
            self._list.LSetCell(self.itemrepr(item), (0, index))
        self.items[index] = item

    def __delitem__(self, index):
        if self._parent and self._list:
            self._list.LDelRow(1, index)
        del self.items[index]

    def __getslice__(self, a, b):
        return self.items[a:b]

    def __delslice__(self, a, b):
        if b-a:
            if self._parent and self._list:
                self._list.LDelRow(b-a, a)
            del self.items[a:b]

    def __setslice__(self, a, b, items):
        if self._parent and self._list:
            l = len(items)
            the_list = self._list
            self.setdrawingmode(0)
            if b-a:
                if b > len(self.items):
                # fix for new 1.5 "feature" where b is sys.maxint instead of len(self)...
                # LDelRow doesn't like maxint.
                    b = len(self.items)
                the_list.LDelRow(b-a, a)
            the_list.LAddRow(l, a)
            self_itemrepr = self.itemrepr
            set_cell = the_list.LSetCell
            for i in range(len(items)):
                set_cell(self_itemrepr(items[i]), (0, i + a))
            self.items[a:b] = items
            self.setdrawingmode(1)
        else:
            self.items[a:b] = items

    def __len__(self):
        return len(self.items)

    def append(self, item):
        if self._parent and self._list:
            index = len(self.items)
            self._list.LAddRow(1, index)
            self._list.LSetCell(self.itemrepr(item), (0, index))
        self.items.append(item)

    def remove(self, item):
        index = self.items.index(item)
        self.__delitem__(index)

    def index(self, item):
        return self.items.index(item)

    def insert(self, index, item):
        if index < 0:
            index = 0
        if self._parent and self._list:
            self._list.LAddRow(1, index)
            self._list.LSetCell(self.itemrepr(item), (0, index))
        self.items.insert(index, item)

    def setdrawingmode(self, onoff):
        if onoff:
            self.drawingmode = self.drawingmode - 1
            if self.drawingmode == 0 and self._list is not None:
                self._list.LSetDrawingMode(1)
                if self._visible:
                    bounds = l, t, r, b = Qd.InsetRect(self._bounds, 1, 1)
                    cl, ct, cr, cb = self._list.LRect((0, len(self.items)-1))
                    if cb < b:
                        self.SetPort()
                        Qd.EraseRect((l, cb, cr, b))
                    self._list.LUpdate(self._parentwindow.wid.GetWindowPort().visRgn)
                    self.GetWindow().ValidWindowRect(bounds)
        else:
            if self.drawingmode == 0 and self._list is not None:
                self._list.LSetDrawingMode(0)
            self.drawingmode = self.drawingmode + 1


class CustomList(List):

    """Base class for writing custom list definitions."""

    _cellHeight = 0

    def createlist(self):
        self._calcbounds()
        self.SetPort()
        rect = self._bounds
        rect = rect[0]+1, rect[1]+1, rect[2]-16, rect[3]-1
        self._viewbounds = rect
        self._list = CreateCustomList(
                        rect,
                        (0, 0, 1, 0),
                        (0, self._cellHeight),
                        (kListDefUserProcType, self.listDefinitionFunc),
                        self._parentwindow.wid,
                        0, 1, 0, 1)
        if self.drawingmode:
            self._list.LSetDrawingMode(0)
        self._list.selFlags = self._flags
        self.setitems(self.items)
        if hasattr(self, "_sel"):
            self.setselection(self._sel)
            del self._sel

    def listDefinitionFunc(self, message, selected, cellRect, theCell,
                    dataOffset, dataLen, theList):
        """The LDEF message dispatcher routine, no need to override."""
        if message == lInitMsg:
            self.listDefInit(theList)
        elif message == lDrawMsg:
            self.listDefDraw(selected, cellRect, theCell,
                            dataOffset, dataLen, theList)
        elif message == lHiliteMsg:
            self.listDefHighlight(selected, cellRect, theCell,
                            dataOffset, dataLen, theList)
        elif message == lCloseMsg:
            self.listDefClose(theList)

    def listDefInit(self, theList):
        pass
    def listDefClose(self, theList):
        pass
    def listDefDraw(self, selected, cellRect, theCell,
                    dataOffset, dataLen, theList):
        pass
    def listDefHighlight(self, selected, cellRect, theCell,
                    dataOffset, dataLen, theList):
        pass


class TwoLineList(CustomList):

    _cellHeight = 28

    def listDefDraw(self, selected, cellRect, theCell,
                    dataOffset, dataLen, theList):
        savedPort = Qd.GetPort()
        Qd.SetPort(theList.GetListPort())
        savedClip = Qd.NewRgn()
        Qd.GetClip(savedClip)
        Qd.ClipRect(cellRect)
        savedPenState = Qd.GetPenState()
        Qd.PenNormal()
        Qd.EraseRect(cellRect)

        #draw the cell if it contains data
        ascent, descent, leading, size, hm = Fm.FontMetrics()
        linefeed = ascent + descent + leading

        if dataLen:
            left, top, right, bottom = cellRect
            data = theList.LGetCell(dataLen, theCell)
            lines = data.split("\r")
            line1 = lines[0]
            if len(lines) > 1:
                line2 = lines[1]
            else:
                line2 = ""
            Qd.MoveTo(int(left + 4), int(top + ascent))
            Qd.DrawText(line1, 0, len(line1))
            if line2:
                Qd.MoveTo(int(left + 4), int(top + ascent + linefeed))
                Qd.DrawText(line2, 0, len(line2))
            Qd.PenPat("\x11\x11\x11\x11\x11\x11\x11\x11")
            bottom = top + theList.cellSize[1]
            Qd.MoveTo(left, bottom - 1)
            Qd.LineTo(right, bottom - 1)
        if selected:
            self.listDefHighlight(selected, cellRect, theCell,
                            dataOffset, dataLen, theList)
        #restore graphics environment
        Qd.SetPort(savedPort)
        Qd.SetClip(savedClip)
        Qd.DisposeRgn(savedClip)
        Qd.SetPenState(savedPenState)

    def listDefHighlight(self, selected, cellRect, theCell,
                    dataOffset, dataLen, theList):
        savedPort = Qd.GetPort()
        Qd.SetPort(theList.GetListPort())
        savedClip = Qd.NewRgn()
        Qd.GetClip(savedClip)
        Qd.ClipRect(cellRect)
        savedPenState = Qd.GetPenState()
        Qd.PenNormal()
        Qd.PenMode(hilitetransfermode)
        Qd.PaintRect(cellRect)

        #restore graphics environment
        Qd.SetPort(savedPort)
        Qd.SetClip(savedClip)
        Qd.DisposeRgn(savedClip)
        Qd.SetPenState(savedPenState)


class ResultsWindow:

    """Simple results window. The user cannot make this window go away completely:
    closing it will just hide it. It will remain in the windows list. The owner of this window
    should call the done() method to indicate it's done with it.
    """

    def __init__(self, possize=None, title="Results", callback=None):
        import W
        if possize is None:
            possize = (500, 200)
        self.w = W.Window(possize, title, minsize=(200, 100))
        self.w.results = W.TwoLineList((-1, -1, 1, -14), callback=None)
        self.w.bind("<close>", self.hide)
        self.w.open()
        self._done = 0

    def done(self):
        self._done = 1
        if not self.w.isvisible():
            self.w.close()

    def hide(self):
        if not self._done:
            self.w.show(0)
            return -1

    def append(self, msg):
        if not self.w.isvisible():
            self.w.show(1)
            self.w.select()
        msg = string.replace(msg, '\n', '\r')
        self.w.results.append(msg)
        self.w.results.setselection([len(self.w.results)-1])

    def __getattr__(self, attr):
        return getattr(self.w.results, attr)


class MultiList(List):

    """XXX Experimantal!!!"""

    def setitems(self, items):
        self.items = items
        if not self._parent or not self._list:
            return
        self._list.LDelRow(0, 1)
        self.setdrawingmode(0)
        self._list.LAddRow(len(self.items), 0)
        self_itemrepr = self.itemrepr
        set_cell = self._list.LSetCell
        for i in range(len(items)):
            row = items[i]
            for j in range(len(row)):
                item = row[j]
                set_cell(self_itemrepr(item), (j, i))
        self.setdrawingmode(1)

    def getselection(self):
        if not self._parent or not self._list:
            if hasattr(self, "_sel"):
                return self._sel
            return []
        items = []
        point = (0,0)
        while 1:
            ok, point = self._list.LGetSelect(1, point)
            if not ok:
                break
            items.append(point[1])
            point = point[0], point[1]+1
        return items

    def setselection(self, selection):
        if not self._parent or not self._list:
            self._sel = selection
            return
        set_sel = self._list.LSetSelect
        for i in range(len(self.items)):
            for j in range(len(self.items[i])):
                if i in selection:
                    set_sel(1, (j, i))
                else:
                    set_sel(0, (j, i))
        #self._list.LAutoScroll()



from Carbon import Ctl, Controls
from Carbon import Evt, Qd, Win
import Wbase


class ControlWidget(Wbase.ClickableWidget):

    """Baseclass for all native controls."""

    def __init__(self, possize, title = "Control", procID = 0, callback = None, value = 0, min = 0, max = 1, viewsize = 0):
        Wbase.ClickableWidget.__init__(self, possize)
        self._control = None
        self._title = title
        self._callback = callback
        self._procID = procID
        self._value = value
        self._min = min
        self._max = max
        self._enabled = 1
        self._viewsize = viewsize

    def open(self):
        self._calcbounds()

        # NewControl doesn't accept 32-bit value, min, or max, so for consistency
        # with the new 32-bit set/get methods, out-of-range values are initially
        # set as zero, followed by a 32-bit set of the actual value.
        # Values not representable in 16 bits will fail on MacOS 8.1, however
        # the vast majority of control usage should still be compatible.
        _value, _min, _max = self._value, self._min, self._max
        if -32768 <= _value <= 32767:
            bigvalue = None
        else:
            bigvalue = _value
            _value = 0
        if -32768 <= _min <= 32767:
            bigmin = None
        else:
            bigmin = _min
            _min = 0
        if -32768 <= _max <= 32767:
            bigmax = None
        else:
            bigmax = _max
            _max = 0
        self._control = Ctl.NewControl(self._parentwindow.wid,
                                        self._bounds,
                                        self._title,
                                        1,
                                        _value,
                                        _min,
                                        _max,
                                        self._procID,
                                        0)
        if bigvalue:
            self._control.SetControl32BitValue(bigvalue)
        if bigmin:
            self._control.SetControl32BitMinimum(bigmin)
        if bigmax:
            self._control.SetControl32BitMaximum(bigmax)
        if self._viewsize:
            try:
                self._control.SetControlViewSize(self._viewsize)
                # Not available in MacOS 8.1, but that's OK since it only affects
                # proportional scrollbars which weren't available in 8.1 either.
            except NotImplementedError:
                pass
        self.enable(self._enabled)

    def adjust(self, oldbounds):
        self.SetPort()
        self._control.HideControl()
        self._control.MoveControl(self._bounds[0], self._bounds[1])
        self._control.SizeControl(self._bounds[2] - self._bounds[0], self._bounds[3] - self._bounds[1])
        if self._visible:
            Qd.EraseRect(self._bounds)
            self._control.ShowControl()
            self.GetWindow().ValidWindowRect(self._bounds)

    def close(self):
        self._control.HideControl()
        self._control = None
        Wbase.ClickableWidget.close(self)

    def enable(self, onoff):
        if self._control and self._enabled <> onoff:
            self._control.HiliteControl((not onoff) and 255)
            self._enabled = onoff

    def show(self, onoff):
        self._visible = onoff
        for w in self._widgets:
            w.show(onoff)
        if onoff:
            self._control.ShowControl()
        else:
            self._control.HideControl()

    def activate(self, onoff):
        self._activated = onoff
        if self._enabled:
            if onoff:
                self._control.ActivateControl()
            else:
                self._control.DeactivateControl()

    def draw(self, visRgn = None):
        if self._visible:
            self._control.Draw1Control()

    def test(self, point):
        if Qd.PtInRect(point, self._bounds) and self._enabled:
            return 1
        #ctltype, control = Ctl.FindControl(point, self._parentwindow.wid)
        #if self._enabled and control == self._control:
        #       return 1

    def click(self, point, modifiers):
        if not self._enabled:
            return
        part = self._control.TrackControl(point)
        if part:
            if self._callback:
                Wbase.CallbackCall(self._callback, 0)

    def settitle(self, title):
        if self._control:
            self._control.SetControlTitle(title)
        self._title = title

    def gettitle(self):
        return self._title

    def set(self, value):
        if self._control:
            if -32768 <= value <= 32767:
                # No 32-bit control support in MacOS 8.1, so use
                # the 16-bit interface when possible.
                self._control.SetControlValue(value)
            else:
                self._control.SetControl32BitValue(value)
        else:
            self._value = value

    def get(self):
        if self._control:
            try:
                return self._control.GetControl32BitValue()
                # No 32-bit control support in MacOS 8.1, so fall
                # back to the 16-bit interface when needed.
            except NotImplementedError:
                return self._control.GetControlValue()
        else:
            return self._value


class Button(ControlWidget):

    """Standard push button."""

    procID = Controls.pushButProc | Controls.useWFont

    def __init__(self, possize, title = "Button", callback = None):
        ControlWidget.__init__(self, possize, title, self.procID, callback, 0, 0, 1)
        self._isdefault = 0

    def push(self):
        if not self._enabled:
            return
        # emulate the pushing of the button
        import time
        self._control.HiliteControl(Controls.kControlButtonPart)
        self._parentwindow.wid.GetWindowPort().QDFlushPortBuffer(None)  # needed under OSX
        time.sleep(0.1)
        self._control.HiliteControl(0)
        if self._callback:
            Wbase.CallbackCall(self._callback, 0)

    def enable(self, onoff):
        if self._control and self._enabled <> onoff:
            self._control.HiliteControl((not onoff) and 255)
            self._enabled = onoff

    def show(self, onoff):
        ControlWidget.show(self, onoff)

    def draw(self, visRgn = None):
        if self._visible:
            self._control.Draw1Control()

    def open(self):
        ControlWidget.open(self)
        if self._isdefault:
            self._setdefault(self._isdefault)

    def _setdefault(self, onoff):
        c = self._control
        if c is not None:
            if onoff:
                data = "\xFF"
            else:
                data = "\0"
            # hide before changing state, otherwise the button isn't always
            # redrawn correctly, although it's quite different under Aqua
            # and Classic...
            c.HideControl()
            c.SetControlData(Controls.kControlNoPart,
                            Controls.kControlPushButtonDefaultTag, data)
            c.ShowControl()
        self._isdefault = onoff

    def adjust(self, oldbounds):
        if self._isdefault:
            old = Qd.InsetRect(oldbounds, -4, -4)
            new = Qd.InsetRect(self._bounds, -4, -4)
            Qd.EraseRect(old)
            self.GetWindow().InvalWindowRect(old)
            self.GetWindow().InvalWindowRect(new)
        ControlWidget.adjust(self, oldbounds)


class BevelButton(Button):
    procID = Controls.kControlBevelButtonNormalBevelProc | Controls.useWFont


class CheckBox(ControlWidget):

    """Standard checkbox."""

    def __init__(self, possize, title = "Checkbox", callback = None, value = 0):
        procID = Controls.checkBoxProc | Controls.useWFont
        ControlWidget.__init__(self, possize, title, procID, callback, value, 0, 1)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        part = self._control.TrackControl(point)
        if part:
            self.toggle()
            if self._callback:
                Wbase.CallbackCall(self._callback, 0, self.get())

    def push(self):
        if not self._enabled:
            return
        self.toggle()
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, self.get())

    def toggle(self):
        self.set(not self.get())


class RadioButton(ControlWidget):

    """Standard radiobutton."""

    # XXX We need a radiogroup widget; this is too kludgy.

    def __init__(self, possize, title, thebuttons, callback = None, value = 0):
        procID = Controls.radioButProc | Controls.useWFont
        ControlWidget.__init__(self, possize, title, procID, callback, value, 0, 1)
        self.thebuttons = thebuttons
        thebuttons.append(self)

    def close(self):
        self.thebuttons = None
        ControlWidget.close(self)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        part = self._control.TrackControl(point)
        if part:
            self.set(1)
            if self._callback:
                Wbase.CallbackCall(self._callback, 0, 1)

    def push(self):
        if not self._enabled:
            return
        self.set(1)
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, 1)

    def set(self, value):
        for button in self.thebuttons:
            if button._control:
                button._control.SetControlValue(button == self)
            else:
                button._value = (button == self)


class Scrollbar(ControlWidget):

    """Standard scrollbar."""

    def __init__(self, possize, callback=None, value=0, min=0, max=0, livefeedback=1):
        if livefeedback:
            procID = Controls.kControlScrollBarLiveProc
        else:
            procID = Controls.scrollBarProc
        ControlWidget.__init__(self, possize, "", procID, callback, value, min, max)

    # interface
#       def set(self, value):
#               if self._callback:
#                       Wbase.CallbackCall(self._callback, 1, value)

    def up(self):
        if self._callback:
            Wbase.CallbackCall(self._callback, 1, '+')

    def down(self):
        if self._callback:
            Wbase.CallbackCall(self._callback, 1, '-')

    def pageup(self):
        if self._callback:
            Wbase.CallbackCall(self._callback, 1, '++')

    def pagedown(self):
        if self._callback:
            Wbase.CallbackCall(self._callback, 1, '--')

    def setmin(self, min):
        if self._control is not None:
            if -32768 <= min <= 32767:
                # No 32-bit control support in MacOS 8.1, so use
                # the 16-bit interface when possible.
                self._control.SetControlMinimum(min)
            else:
                self._control.SetControl32BitMinimum(min)
        else:
            self._min = min

    def setmax(self, max):
        if self._control is not None:
            if -32768 <= max <= 32767:
                # No 32-bit control support in MacOS 8.1, so use
                # the 16-bit interface when possible.
                self._control.SetControlMaximum(max)
            else:
                self._control.SetControl32BitMaximum(max)
        else:
            self._max = max

    def setviewsize(self, viewsize):
        if self._control is not None:
            try:
                self._control.SetControlViewSize(viewsize)
                # Not available in MacOS 8.1, but that's OK since it only affects
                # proportional scrollbars which weren't available in 8.1 either.
            except NotImplementedError:
                pass
        else:
            self._viewsize = viewsize

    def getmin(self):
        try:
            return self._control.GetControl32BitMinimum()
            # No 32-bit control support in MacOS 8.1, so fall
            # back to the 16-bit interface when needed.
        except NotImplementedError:
            return self._control.GetControlMinimum()

    def getmax(self):
        try:
            return self._control.GetControl32BitMaximum()
            # No 32-bit control support in MacOS 8.1, so fall
            # back to the 16-bit interface when needed.
        except NotImplementedError:
            return self._control.GetControlMaximum()

    # internals
    def click(self, point, modifiers):
        if not self._enabled:
            return
        def hitter(ctl, part, self=self):
            if part:
                self._hit(part)
        part = self._control.TrackControl(point, hitter)

    def _hit(self, part):
        value = None
        if part == Controls.inThumb:
            try:
                value = self._control.GetControl32BitValue()
                # No 32-bit control support in MacOS 8.1, so fall
                # back to the 16-bit interface when needed.
            except NotImplementedError:
                value = self._control.GetControlValue()
        elif part == Controls.inUpButton:
            value = "+"
        elif part == Controls.inDownButton:
            value = "-"
        elif part == Controls.inPageUp:
            value = "++"
        elif part == Controls.inPageDown:
            value = "--"
        if value is not None and self._callback:
            Wbase.CallbackCall(self._callback, 1, value)

    def draw(self, visRgn = None):
        if self._visible:
            self._control.Draw1Control()
            #Qd.FrameRect(self._bounds)

    def adjust(self, oldbounds):
        self.SetPort()
        self.GetWindow().InvalWindowRect(oldbounds)
        self._control.HideControl()
        self._control.MoveControl(self._bounds[0], self._bounds[1])
        self._control.SizeControl(self._bounds[2] - self._bounds[0], self._bounds[3] - self._bounds[1])
        if self._visible:
            Qd.EraseRect(self._bounds)
            if self._activated:
                self._control.ShowControl()
            else:
                Qd.FrameRect(self._bounds)
            self.GetWindow().ValidWindowRect(self._bounds)


def _scalebarvalue(absmin, absmax, curmin, curmax):
    if curmin <= absmin and curmax >= absmax:
        return None
    if curmin <= absmin:
        return 0
    if curmax >= absmax:
        return 32767
    perc = float(curmin-absmin) / float((absmax - absmin) - (curmax - curmin))
    return int(perc*32767)



import re
import os

# The regular expression for freeze directives. These are comments with the
# word macfreeze immedeately followed by a colon, followed by a directive,
# followed by argument(s)
#
# The directives supported are
# include - Include a module or file
# exclude - Exclude a module
# optional - Include a module if it is found, but don't complain if it isn't
# path - Add sys.path entries. Relative paths are relative to the source file.
#
# See the macfreeze.py main program for a real live example.
#
DIRECTIVE_RE=r'^\s*#\s*macfreeze:\s*(\S*)\s*(.*)\s*$'
REPROG=re.compile(DIRECTIVE_RE)

def findfreezedirectives(program):
    extra_modules = []
    exclude_modules = []
    optional_modules = []
    extra_path = []
    progdir, filename = os.path.split(program)
    fp = open(program)
    for line in fp.readlines():
        match = REPROG.match(line)
        if match:
            directive = match.group(1)
            argument = match.group(2)
            if directive == 'include':
                extra_modules.append(argument)
            elif directive == 'exclude':
                exclude_modules.append(argument)
            elif directive == 'optional':
                optional_modules.append(argument)
            elif directive == 'path':
                argument = os.path.join(progdir, argument)
                extra_path.append(argument)
            else:
                print '** Unknown directive', line
    return extra_modules, exclude_modules, optional_modules, extra_path
